#!/usr/bin/python
# -*- coding: utf-8 -*-
"""unittests for acme2certifier"""

# pylint: disable= C0415, W0212
import unittest
import sys
import os
from unittest.mock import patch, Mock, MagicMock
import base64
import configparser

sys.path.insert(0, ".")
sys.path.insert(1, "..")


class TestACMEHandler(unittest.TestCase):
    """test class for cgi_handler"""

    def setUp(self):
        """setup unittest"""
        import logging

        logging.basicConfig(level=logging.CRITICAL)
        self.logger = logging.getLogger("test_a2c")
        from acme2certifier.cahandlers.mscertsrv_ca_handler import CAhandler

        self.cahandler = CAhandler(False, self.logger)
        self.dir_path = os.path.dirname(os.path.realpath(__file__))

    def test_001_default(self):
        """default test which always passes"""
        self.assertEqual("foo", "foo")

    def test_002__pkcs7_to_pem(self):
        """test pkcs7 to pem default output"""
        with open(self.dir_path + "/ca/certs.p7b", "r") as fso:
            file_content = fso.read()
        with open(self.dir_path + "/ca/certs.pem", "r") as fso:
            result = fso.read()
        self.assertEqual(result, self.cahandler._pkcs7_to_pem(file_content))

    def test_003__pkcs7_to_pem(self):
        """test pkcs7 to pem output string"""
        with open(self.dir_path + "/ca/certs.p7b", "r") as fso:
            file_content = fso.read()
        with open(self.dir_path + "/ca/certs.pem", "r") as fso:
            result = fso.read()
        self.assertEqual(result, self.cahandler._pkcs7_to_pem(file_content, "string"))

    def test_004__pkcs7_to_pem(self):
        """test pkcs7 to pem output list"""
        with open(self.dir_path + "/ca/certs.p7b", "r") as fso:
            file_content = fso.read()
        result = [
            "-----BEGIN CERTIFICATE-----\nMIIFTzCCAzegAwIBAgIIAzHyhSyrXfMwDQYJKoZIhvcNAQELBQAwKzEXMBUGA1UE\nCxMOYWNtZTJjZXJ0aWZpZXIxEDAOBgNVBAMTB3Jvb3QtY2EwHhcNMjAwNTI3MTM1\nNDAwWhcNMzAwNTI2MjM1OTAwWjAqMRcwFQYDVQQLEw5hY21lMmNlcnRpZmllcjEP\nMA0GA1UEAxMGc3ViLWNhMIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEA\nxXHaGZsolXe+PBdUryngHP9VbBC1mehqeTtYI+hqsqGNH7q9a7bSrxMwFuF1kYL8\njqqxkJdtl0L94xcxJg/ZdMx7Nt0vGI+BaAuTpEpUEHeN4tqS6NhB/m/0LGkAELc/\nqkzmoO4B1FDwEEj/3IXtZcupqG80oDt7jWSGXdtF7NTjzcumznMeRXidCdhxRxT/\n/WrsChaytXo0xWZ56oeNwd6x6Dr8/39PBOWtj4fldyDcg+Q+alci2tx9pxmu2bCV\nXcB9ftCLKhDk2WEHE88bgKSp7fV2RCmq9po+Tx8JJ7qecLunUsK/F0XN4kpoQLm9\nhcymqchnMSncSiyin1dQHGHWgXDtBDdq6A2Z6rx26Qk5H9HTYvcNSe1YwFEDoGLB\nZQjbCPWiaqoaH4agBQTclPvrrSCRaVmhUSO+pBtSXDkmN4t3MDZxfgRkp8ixwkB1\n5Y5f0LTpCyAJsdQDw8+Ea0aDqO30eskh4CErnm9+Fejd9Ew2cwpdwfBXzVSbYilM\nGueQihZHvJmVRxAwU69aO2Qs8B0tQ60CfWKVlmWPiakrvYYlPp0FBsM61G6LZEN8\nhH2CKnS8hHv5IWEXZvp0Pk8V3P5h6bWN0Tl+x/V1Prt7Wp8NoiPETE8XyDDxe6dm\nKxztWBH/mTsJyMGb6ZiUoXdPU9TFUKqHxTRLHaxfsPsCAwEAAaN4MHYwEgYDVR0T\nAQH/BAgwBgEB/wIBATAdBgNVHQ4EFgQUv96OjgYiIqutQ8jd1E+oq0hBPtUwDgYD\nVR0PAQH/BAQDAgGGMBEGCWCGSAGG+EIBAQQEAwIABzAeBglghkgBhvhCAQ0EERYP\neGNhIGNlcnRpZmljYXRlMA0GCSqGSIb3DQEBCwUAA4ICAQBbHLEVyg4f9uEujroc\n31UVyDRLMdPgEPLjOenSBCBmH0N81whDmxNI/7JAAB6J14WMX8OLF0HkZnb7G77W\nvDhy1aFvQFbXHBz3/zUO9Mw9J4L2XEW6ond3Nsh1m2oXeBde3R3ANxuIzHqZDlP9\n6YrRcHjnf4+1/5AKDJAvJD+gFb5YnYUKH2iSvHUvG17xcZx98Rf2eo8LealG4JqH\nJh4sKRy0VjDQD7jXSCbweTHEb8wz+6OfNGrIo+BhTFP5vPcwE4nlJwYBoaOJ5cVa\n7gdQJ7WkLSxvwHxuxzvSVK73u3jl3I9SqTrbMLG/jeJyV0P8EvdljOaGnCtQVRwC\nzM4ptXUvKhKOHy7/nyTF/Bc35ZwwL/2xWvNK1+NibgE/6CFxupwWpdmxQbVVuoQ3\n2tUil9ty0yC6m5GKE8+t1lrZuxyA+b/TBnYNO5xo8UEMbkpxaNYSwmw+f/loxXP/\nM7sIBcLvy2ugHEBxwd9o/kLXeXT2DaRvxPjp4yk8MpJRpNmz3aB5HJwaUnaRLVo5\nZ3XWWXmjMGZ6/m0AAoDbDz/pXtOoJZT8BJdD1DuDdszVsQnLVn4B/LtIXL6FbXsF\nzfv6ERP9a5gpKUZ+4NjgrnlGtdccNZpwyWF0IXcvaq3b8hXIRO4hMjzHeHfzJN4t\njX1vlY35Ofonc4+6dRVamBiF9A==\n-----END CERTIFICATE-----\n",
            "-----BEGIN CERTIFICATE-----\nMIIFcDCCA1igAwIBAgIIevLTTxOMoZgwDQYJKoZIhvcNAQELBQAwKzEXMBUGA1UE\nCxMOYWNtZTJjZXJ0aWZpZXIxEDAOBgNVBAMTB3Jvb3QtY2EwHhcNMjAwNTI3MDAw\nMDAwWhcNMzAwNTI2MjM1OTU5WjArMRcwFQYDVQQLEw5hY21lMmNlcnRpZmllcjEQ\nMA4GA1UEAxMHcm9vdC1jYTCCAiIwDQYJKoZIhvcNAQEBBQADggIPADCCAgoCggIB\nAJy4UZHdZgYt64k/rFamoC676tYvtabeuiqVw1c6oVZI897cFLG6BYwyr2Eaj7tF\nrqTJDeMN4vZSudLsmLDq6m8KwX/riPzUTIlcjM5aIMANZr9rLEs3NWtcivolB5aQ\n1slhdVitUPLuxsFnYeQTyxFyP7lng9M/Z403KLG8phdmKjM0vJkaj4OuKOXf3UsW\nqWQYyRl/ms07xVj02uq08LkoeO+jtQisvyVXURdaCceZtyK/ZBQ7NFCsbK112cVR\n1e2aJol7NJAA6Wm6iBzAdkAA2l3kh40SLoEbaiaVMixLN2vilIZOOAoDXX4+T6ir\n+KnDVSJ2yu5c/OJMwuXwHrh7Lgg1vsFR5TNehknhjUuWOUO+0TkKPg2A7KTg72OZ\n2mOcLZIbxzr1P5RRvdmLQLPrTF2EJvpQPNmbXqN3ZVWEvfHTjkkTFY/dsOTvFTgS\nri15zYKch8votcU7z+BQhgmMtwO2JhPMmZ6ABd9skI7ijWpwOltAhxtdoBO6T6CB\nCrE2yXc6V/PyyAKcFglNmIght5oXsnE+ub/dtx8f9Iea/xNPdo5aGy8fdaitolDK\n16kd3Kb7OE4HMHIwOxxF1BEAqerxxhbLMRBr8hRSZI5cvLzWLvpAQ5zuhjD6V3b9\nBYFd4ujAu3zl3mbzdbYjFoGOX6aBZaGDxlc4O2W7HxntAgMBAAGjgZcwgZQwDwYD\nVR0TAQH/BAUwAwEB/zAdBgNVHQ4EFgQUDGVvuTFYZtEAkz3af9wRKDDvAswwHwYD\nVR0jBBgwFoAUDGVvuTFYZtEAkz3af9wRKDDvAswwDgYDVR0PAQH/BAQDAgGGMBEG\nCWCGSAGG+EIBAQQEAwIABzAeBglghkgBhvhCAQ0EERYPeGNhIGNlcnRpZmljYXRl\nMA0GCSqGSIb3DQEBCwUAA4ICAQAjko7dX+iCgT+m3Iy1Vg6j7MRevPAzq1lqHRRN\nNdt2ct530pIut7Fv5V2xYk35ka+i/G+XyOvTXa9vAUKiBtiRnUPsXu4UcS7CcrCX\nEzHx4eOtHnp5wDhO0Fx5/OUZTaP+L7Pd1GD/j953ibx5bMa/M9Rj+S486nst57tu\nDRmEAavFDiMd6L3jH4YSckjmIH2uSeDIaRa9k6ag077XmWhvVYQ9tuR7RGbSuuV3\nFc6pqcFbbWpoLhNRcFc+hbUKOsKl2cP+QEKP/H2s3WMllqgAKKZeO+1KOsGo1CDs\n475bIXyCBpFbH2HOPatmu3yZRQ9fj9ta9EW46n33DFRNLinFWa4WJs4yLVP1juge\n2TCOyA1t61iy++RRXSG3e7NFYrEZuCht1EdDAdzIUY89m9NCPwoDYS4CahgnfkkO\n7YQe6f6yqK6isyf8ZFcp1uF58eERDiF/FDqS8nLmCdURuI56DDoNvDpig5J/9RNW\nG8vEvt2p7QrjeZ3EAatx5JuYty/NKTHZwJWk51CgzEgzDwzE2JIiqeldtL5d0Sl6\neVuv0G04BEyuXxEWpgVVzBS4qEFIBSnTJzgu1PXmId3yLvg2Nr8NKvwyZmN5xKFp\n0A9BWo15zW1PXDaD+l39oTYD7agjXkzTAjYIcfNJ7ATIYFD0xAvNAOf70s7aNupF\nfvkG2Q==\n-----END CERTIFICATE-----\n",
        ]
        self.assertEqual(result, self.cahandler._pkcs7_to_pem(file_content, "list"))

    def test_005__pkcs7_to_pem(self):
        """test pkcs7 to pem output list"""
        with open(self.dir_path + "/ca/certs.p7b", "r") as fso:
            file_content = fso.read()
        result = None
        self.assertEqual(result, self.cahandler._pkcs7_to_pem(file_content, "unknown"))

    def test_006__pkcs7_to_pem(self):
        """test pkcs7 to pem output list"""

        file_content = base64.b64decode(
            "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"
        )
        result = [
            "-----BEGIN CERTIFICATE-----\nMIIFTzCCAzegAwIBAgIIAzHyhSyrXfMwDQYJKoZIhvcNAQELBQAwKzEXMBUGA1UE\nCxMOYWNtZTJjZXJ0aWZpZXIxEDAOBgNVBAMTB3Jvb3QtY2EwHhcNMjAwNTI3MTM1\nNDAwWhcNMzAwNTI2MjM1OTAwWjAqMRcwFQYDVQQLEw5hY21lMmNlcnRpZmllcjEP\nMA0GA1UEAxMGc3ViLWNhMIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEA\nxXHaGZsolXe+PBdUryngHP9VbBC1mehqeTtYI+hqsqGNH7q9a7bSrxMwFuF1kYL8\njqqxkJdtl0L94xcxJg/ZdMx7Nt0vGI+BaAuTpEpUEHeN4tqS6NhB/m/0LGkAELc/\nqkzmoO4B1FDwEEj/3IXtZcupqG80oDt7jWSGXdtF7NTjzcumznMeRXidCdhxRxT/\n/WrsChaytXo0xWZ56oeNwd6x6Dr8/39PBOWtj4fldyDcg+Q+alci2tx9pxmu2bCV\nXcB9ftCLKhDk2WEHE88bgKSp7fV2RCmq9po+Tx8JJ7qecLunUsK/F0XN4kpoQLm9\nhcymqchnMSncSiyin1dQHGHWgXDtBDdq6A2Z6rx26Qk5H9HTYvcNSe1YwFEDoGLB\nZQjbCPWiaqoaH4agBQTclPvrrSCRaVmhUSO+pBtSXDkmN4t3MDZxfgRkp8ixwkB1\n5Y5f0LTpCyAJsdQDw8+Ea0aDqO30eskh4CErnm9+Fejd9Ew2cwpdwfBXzVSbYilM\nGueQihZHvJmVRxAwU69aO2Qs8B0tQ60CfWKVlmWPiakrvYYlPp0FBsM61G6LZEN8\nhH2CKnS8hHv5IWEXZvp0Pk8V3P5h6bWN0Tl+x/V1Prt7Wp8NoiPETE8XyDDxe6dm\nKxztWBH/mTsJyMGb6ZiUoXdPU9TFUKqHxTRLHaxfsPsCAwEAAaN4MHYwEgYDVR0T\nAQH/BAgwBgEB/wIBATAdBgNVHQ4EFgQUv96OjgYiIqutQ8jd1E+oq0hBPtUwDgYD\nVR0PAQH/BAQDAgGGMBEGCWCGSAGG+EIBAQQEAwIABzAeBglghkgBhvhCAQ0EERYP\neGNhIGNlcnRpZmljYXRlMA0GCSqGSIb3DQEBCwUAA4ICAQBbHLEVyg4f9uEujroc\n31UVyDRLMdPgEPLjOenSBCBmH0N81whDmxNI/7JAAB6J14WMX8OLF0HkZnb7G77W\nvDhy1aFvQFbXHBz3/zUO9Mw9J4L2XEW6ond3Nsh1m2oXeBde3R3ANxuIzHqZDlP9\n6YrRcHjnf4+1/5AKDJAvJD+gFb5YnYUKH2iSvHUvG17xcZx98Rf2eo8LealG4JqH\nJh4sKRy0VjDQD7jXSCbweTHEb8wz+6OfNGrIo+BhTFP5vPcwE4nlJwYBoaOJ5cVa\n7gdQJ7WkLSxvwHxuxzvSVK73u3jl3I9SqTrbMLG/jeJyV0P8EvdljOaGnCtQVRwC\nzM4ptXUvKhKOHy7/nyTF/Bc35ZwwL/2xWvNK1+NibgE/6CFxupwWpdmxQbVVuoQ3\n2tUil9ty0yC6m5GKE8+t1lrZuxyA+b/TBnYNO5xo8UEMbkpxaNYSwmw+f/loxXP/\nM7sIBcLvy2ugHEBxwd9o/kLXeXT2DaRvxPjp4yk8MpJRpNmz3aB5HJwaUnaRLVo5\nZ3XWWXmjMGZ6/m0AAoDbDz/pXtOoJZT8BJdD1DuDdszVsQnLVn4B/LtIXL6FbXsF\nzfv6ERP9a5gpKUZ+4NjgrnlGtdccNZpwyWF0IXcvaq3b8hXIRO4hMjzHeHfzJN4t\njX1vlY35Ofonc4+6dRVamBiF9A==\n-----END CERTIFICATE-----\n",
            "-----BEGIN CERTIFICATE-----\nMIIFcDCCA1igAwIBAgIIevLTTxOMoZgwDQYJKoZIhvcNAQELBQAwKzEXMBUGA1UE\nCxMOYWNtZTJjZXJ0aWZpZXIxEDAOBgNVBAMTB3Jvb3QtY2EwHhcNMjAwNTI3MDAw\nMDAwWhcNMzAwNTI2MjM1OTU5WjArMRcwFQYDVQQLEw5hY21lMmNlcnRpZmllcjEQ\nMA4GA1UEAxMHcm9vdC1jYTCCAiIwDQYJKoZIhvcNAQEBBQADggIPADCCAgoCggIB\nAJy4UZHdZgYt64k/rFamoC676tYvtabeuiqVw1c6oVZI897cFLG6BYwyr2Eaj7tF\nrqTJDeMN4vZSudLsmLDq6m8KwX/riPzUTIlcjM5aIMANZr9rLEs3NWtcivolB5aQ\n1slhdVitUPLuxsFnYeQTyxFyP7lng9M/Z403KLG8phdmKjM0vJkaj4OuKOXf3UsW\nqWQYyRl/ms07xVj02uq08LkoeO+jtQisvyVXURdaCceZtyK/ZBQ7NFCsbK112cVR\n1e2aJol7NJAA6Wm6iBzAdkAA2l3kh40SLoEbaiaVMixLN2vilIZOOAoDXX4+T6ir\n+KnDVSJ2yu5c/OJMwuXwHrh7Lgg1vsFR5TNehknhjUuWOUO+0TkKPg2A7KTg72OZ\n2mOcLZIbxzr1P5RRvdmLQLPrTF2EJvpQPNmbXqN3ZVWEvfHTjkkTFY/dsOTvFTgS\nri15zYKch8votcU7z+BQhgmMtwO2JhPMmZ6ABd9skI7ijWpwOltAhxtdoBO6T6CB\nCrE2yXc6V/PyyAKcFglNmIght5oXsnE+ub/dtx8f9Iea/xNPdo5aGy8fdaitolDK\n16kd3Kb7OE4HMHIwOxxF1BEAqerxxhbLMRBr8hRSZI5cvLzWLvpAQ5zuhjD6V3b9\nBYFd4ujAu3zl3mbzdbYjFoGOX6aBZaGDxlc4O2W7HxntAgMBAAGjgZcwgZQwDwYD\nVR0TAQH/BAUwAwEB/zAdBgNVHQ4EFgQUDGVvuTFYZtEAkz3af9wRKDDvAswwHwYD\nVR0jBBgwFoAUDGVvuTFYZtEAkz3af9wRKDDvAswwDgYDVR0PAQH/BAQDAgGGMBEG\nCWCGSAGG+EIBAQQEAwIABzAeBglghkgBhvhCAQ0EERYPeGNhIGNlcnRpZmljYXRl\nMA0GCSqGSIb3DQEBCwUAA4ICAQAjko7dX+iCgT+m3Iy1Vg6j7MRevPAzq1lqHRRN\nNdt2ct530pIut7Fv5V2xYk35ka+i/G+XyOvTXa9vAUKiBtiRnUPsXu4UcS7CcrCX\nEzHx4eOtHnp5wDhO0Fx5/OUZTaP+L7Pd1GD/j953ibx5bMa/M9Rj+S486nst57tu\nDRmEAavFDiMd6L3jH4YSckjmIH2uSeDIaRa9k6ag077XmWhvVYQ9tuR7RGbSuuV3\nFc6pqcFbbWpoLhNRcFc+hbUKOsKl2cP+QEKP/H2s3WMllqgAKKZeO+1KOsGo1CDs\n475bIXyCBpFbH2HOPatmu3yZRQ9fj9ta9EW46n33DFRNLinFWa4WJs4yLVP1juge\n2TCOyA1t61iy++RRXSG3e7NFYrEZuCht1EdDAdzIUY89m9NCPwoDYS4CahgnfkkO\n7YQe6f6yqK6isyf8ZFcp1uF58eERDiF/FDqS8nLmCdURuI56DDoNvDpig5J/9RNW\nG8vEvt2p7QrjeZ3EAatx5JuYty/NKTHZwJWk51CgzEgzDwzE2JIiqeldtL5d0Sl6\neVuv0G04BEyuXxEWpgVVzBS4qEFIBSnTJzgu1PXmId3yLvg2Nr8NKvwyZmN5xKFp\n0A9BWo15zW1PXDaD+l39oTYD7agjXkzTAjYIcfNJ7ATIYFD0xAvNAOf70s7aNupF\nfvkG2Q==\n-----END CERTIFICATE-----\n",
        ]
        self.assertEqual(result, self.cahandler._pkcs7_to_pem(file_content, "list"))

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_007_config_load(self, mock_load_cfg):
        """test _config_load no cahandler section"""
        parser = configparser.ConfigParser()
        # parser['CAhandler'] = {'foo': 'bar'}
        self.cahandler._config_load()
        self.assertFalse(self.cahandler.host)
        self.assertFalse(self.cahandler.user)
        self.assertFalse(self.cahandler.password)
        self.assertEqual("basic", self.cahandler.auth_method)
        self.assertFalse(self.cahandler.ca_bundle)
        self.assertFalse(self.cahandler.template)
        self.assertFalse(self.cahandler.krb5_config)

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_008_config_load(self, mock_load_cfg):
        """test _config_load cahandler section with unknown values"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"foo": "bar"}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertFalse(self.cahandler.host)
        self.assertFalse(self.cahandler.user)
        self.assertFalse(self.cahandler.password)
        self.assertEqual("basic", self.cahandler.auth_method)
        self.assertFalse(self.cahandler.ca_bundle)
        self.assertFalse(self.cahandler.template)
        self.assertFalse(self.cahandler.krb5_config)

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_009_config_load(self, mock_load_cfg):
        """test _config_load no cahandler section with host value"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"host": "host"}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertEqual("host", self.cahandler.host)
        self.assertFalse(self.cahandler.user)
        self.assertFalse(self.cahandler.password)
        self.assertEqual("basic", self.cahandler.auth_method)
        self.assertFalse(self.cahandler.ca_bundle)
        self.assertFalse(self.cahandler.template)
        self.assertFalse(self.cahandler.krb5_config)

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_010_config_load(self, mock_load_cfg):
        """test _config_load cahandler section with user values"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"user": "user"}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertFalse(self.cahandler.host)
        self.assertEqual("user", self.cahandler.user)
        self.assertFalse(self.cahandler.password)
        self.assertEqual("basic", self.cahandler.auth_method)
        self.assertFalse(self.cahandler.ca_bundle)
        self.assertFalse(self.cahandler.template)
        self.assertFalse(self.cahandler.krb5_config)

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_011_config_load(self, mock_load_cfg):
        """test _config_load cahandler section with password values"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"password": "password"}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertFalse(self.cahandler.host)
        self.assertFalse(self.cahandler.user)
        self.assertEqual("password", self.cahandler.password)
        self.assertEqual("basic", self.cahandler.auth_method)
        self.assertFalse(self.cahandler.ca_bundle)
        self.assertFalse(self.cahandler.template)
        self.assertFalse(self.cahandler.krb5_config)

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_012_config_load(self, mock_load_cfg):
        """test _config_load cahandler section with authmethod basic"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"auth_method": "basic"}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertFalse(self.cahandler.host)
        self.assertFalse(self.cahandler.user)
        self.assertFalse(self.cahandler.password)
        self.assertEqual("basic", self.cahandler.auth_method)
        self.assertFalse(self.cahandler.ca_bundle)
        self.assertFalse(self.cahandler.template)
        self.assertFalse(self.cahandler.krb5_config)

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_013_config_load(self, mock_load_cfg):
        """test _config_load cahandler section with authmethod ntlm"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"auth_method": "ntlm"}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertFalse(self.cahandler.host)
        self.assertFalse(self.cahandler.user)
        self.assertFalse(self.cahandler.password)
        self.assertEqual("ntlm", self.cahandler.auth_method)
        self.assertFalse(self.cahandler.ca_bundle)
        self.assertFalse(self.cahandler.template)
        self.assertFalse(self.cahandler.krb5_config)

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_013a_config_load_warn_deprecated_basic(self, mock_load_cfg):
        """test warning when auth_method basic is explicitly configured"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"auth_method": "basic"}
        mock_load_cfg.return_value = parser

        with self.assertLogs("test_a2c", level="WARNING") as lcm:
            self.cahandler._config_load()

        self.assertIn(
            "WARNING:test_a2c:Auth method 'basic' is deprecated and will be removed in a future release. Please migrate to 'gssapi' (Kerberos).",
            lcm.output,
        )

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_013b_config_load_warn_deprecated_ntlm(self, mock_load_cfg):
        """test warning when auth_method ntlm is explicitly configured"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"auth_method": "ntlm"}
        mock_load_cfg.return_value = parser

        with self.assertLogs("test_a2c", level="WARNING") as lcm:
            self.cahandler._config_load()

        self.assertIn(
            "WARNING:test_a2c:Auth method 'ntlm' is deprecated and will be removed in a future release. Please migrate to 'gssapi' (Kerberos).",
            lcm.output,
        )

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_014_config_load(self, mock_load_cfg):
        """test _config_load cahandler section with authmethod unknown"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"auth_method": "unknown"}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertFalse(self.cahandler.host)
        self.assertFalse(self.cahandler.user)
        self.assertFalse(self.cahandler.password)
        self.assertEqual("basic", self.cahandler.auth_method)
        self.assertFalse(self.cahandler.ca_bundle)
        self.assertFalse(self.cahandler.template)
        self.assertFalse(self.cahandler.krb5_config)

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_015_config_load(self, mock_load_cfg):
        """test _config_load cahandler section with ca_bundle value"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"ca_bundle": "ca_bundle"}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertFalse(self.cahandler.host)
        self.assertFalse(self.cahandler.user)
        self.assertFalse(self.cahandler.password)
        self.assertEqual("basic", self.cahandler.auth_method)
        self.assertEqual("ca_bundle", self.cahandler.ca_bundle)
        self.assertFalse(self.cahandler.template)
        self.assertFalse(self.cahandler.krb5_config)

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_016_config_load(self, mock_load_cfg):
        """test _config_load cahandler section with template value"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"template": "template"}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertFalse(self.cahandler.host)
        self.assertFalse(self.cahandler.user)
        self.assertFalse(self.cahandler.password)
        self.assertEqual("basic", self.cahandler.auth_method)
        self.assertFalse(self.cahandler.ca_bundle)
        self.assertEqual("template", self.cahandler.template)
        self.assertFalse(self.cahandler.krb5_config)

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_017_config_load(self, mock_load_cfg):
        """test _config_load cahandler section with template value"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"krb5_config": "krb5_config"}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertFalse(self.cahandler.host)
        self.assertFalse(self.cahandler.user)
        self.assertFalse(self.cahandler.password)
        self.assertEqual("basic", self.cahandler.auth_method)
        self.assertFalse(self.cahandler.ca_bundle)
        self.assertFalse(self.cahandler.template)
        self.assertEqual("krb5_config", self.cahandler.krb5_config)

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_017a_config_load_kerberos_keytab_options(self, mock_load_cfg):
        """test _config_load with kerberos keytab related options"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {
            "auth_method": "gssapi",
            "krb5_principal": "svc-a2c-enroll@EXAMPLE.COM",
            "krb5_keytab": "/tmp/svc.keytab",
            "krb5_cache": "/tmp/krb5cc_svc",
            "krb5_config": "/tmp/krb5.conf",
            "krb5_kinit_path": "/usr/local/bin/kinit",
        }
        mock_load_cfg.return_value = parser

        self.cahandler._config_load()

        self.assertEqual("gssapi", self.cahandler.auth_method)
        self.assertEqual("svc-a2c-enroll@EXAMPLE.COM", self.cahandler.krb5_principal)
        self.assertEqual("/tmp/svc.keytab", self.cahandler.krb5_keytab)
        self.assertEqual("/tmp/krb5cc_svc", self.cahandler.krb5_cache)
        self.assertEqual("/tmp/krb5.conf", self.cahandler.krb5_config)
        self.assertEqual("/usr/local/bin/kinit", self.cahandler.krb5_kinit_path)

    @patch.dict("os.environ", {"host_variable": "host"})
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_018_config_load(self, mock_load_cfg):
        """test _config_load - load with host variable"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"host_variable": "host_variable"}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertEqual("host", self.cahandler.host)

    @patch.dict("os.environ", {"host_variable": "host"})
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_019_config_load(self, mock_load_cfg):
        """test _config_load - load with host variable which does not exist"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"host_variable": "doesnotexist"}
        mock_load_cfg.return_value = parser
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertFalse(self.cahandler.host)
        self.assertIn(
            "ERROR:test_a2c:Could not load host_variable from environment: 'doesnotexist'",
            lcm.output,
        )

    @patch.dict("os.environ", {"host_variable": "host"})
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_020_config_load(self, mock_load_cfg):
        """test _config_load - load with host variable which gets overwritten"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"host_variable": "host_variable", "host": "host_local"}
        mock_load_cfg.return_value = parser
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertEqual("host_local", self.cahandler.host)
        self.assertIn("INFO:test_a2c:Overwrite host", lcm.output)

    @patch.dict("os.environ", {"user_variable": "user"})
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_021_config_load(self, mock_load_cfg):
        """test _config_load - load with user variable"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"user_variable": "user_variable"}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertEqual("user", self.cahandler.user)

    @patch.dict("os.environ", {"user_variable": "user"})
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_022_config_load(self, mock_load_cfg):
        """test _config_load - load with user variable which does not exist"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"user_variable": "doesnotexist"}
        mock_load_cfg.return_value = parser
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertFalse(self.cahandler.user)
        self.assertIn(
            "ERROR:test_a2c:Could not load user_variable from environment: 'doesnotexist'",
            lcm.output,
        )

    @patch.dict("os.environ", {"user_variable": "user"})
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_023_config_load(self, mock_load_cfg):
        """test _config_load - load with user variable which gets overwritten"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"user_variable": "user_variable", "user": "user_local"}
        mock_load_cfg.return_value = parser
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertEqual("user_local", self.cahandler.user)
        self.assertIn("INFO:test_a2c:Overwrite user", lcm.output)

    @patch.dict("os.environ", {"password_variable": "password"})
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_024_config_load(self, mock_load_cfg):
        """test _config_load - load with password variable"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"password_variable": "password_variable"}
        mock_load_cfg.return_value = parser
        self.cahandler._config_load()
        self.assertEqual("password", self.cahandler.password)

    @patch.dict("os.environ", {"password_variable": "password"})
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_025_config_load(self, mock_load_cfg):
        """test _config_load - load with password variable which does not exist"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"password_variable": "doesnotexist"}
        mock_load_cfg.return_value = parser
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertFalse(self.cahandler.password)
        self.assertIn(
            "ERROR:test_a2c:Could not load password_variable from environment: 'doesnotexist'",
            lcm.output,
        )

    @patch.dict("os.environ", {"password_variable": "password"})
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_026_config_load(self, mock_load_cfg):
        """test _config_load - load with password variable which gets overwritten"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {
            "password_variable": "password_variable",
            "password": "password_local",
        }
        mock_load_cfg.return_value = parser
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertEqual("password_local", self.cahandler.password)
        self.assertIn("INFO:test_a2c:Overwrite password", lcm.output)

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.proxy_check")
    @patch("json.loads")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_027_config_load(self, mock_load_cfg, mock_json, mock_chk):
        """test _config_load ca_handler configured load proxies"""
        parser = configparser.ConfigParser()
        parser["DEFAULT"] = {"proxy_server_list": "foo"}
        mock_load_cfg.return_value = parser
        mock_json.return_value = "foo.bar.local"
        mock_chk.return_value = "proxy.bar.local"
        self.cahandler._config_load()
        self.assertTrue(mock_json.called)
        self.assertTrue(mock_chk.called)
        self.assertEqual(
            {"http": "proxy.bar.local", "https": "proxy.bar.local"},
            self.cahandler.proxy,
        )

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.proxy_check")
    @patch("json.loads")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_028_config_load(self, mock_load_cfg, mock_json, mock_chk):
        """test _config_load ca_handler configured load proxies failed with exception in json.load"""
        parser = configparser.ConfigParser()
        parser["DEFAULT"] = {"proxy_server_list": "foo"}
        mock_json.side_effect = Exception("exc_load_config")
        mock_load_cfg.return_value = parser
        mock_chk.side = "proxy.bar.local"
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertTrue(mock_json.called)
        self.assertFalse(mock_chk.called)
        self.assertFalse(self.cahandler.proxy)
        self.assertIn(
            "WARNING:test_a2c:Failed to load proxy_server_list from configuration: exc_load_config",
            lcm.output,
        )

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.config_eab_profile_load")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.load_config")
    def test_029_config_load(self, mock_load_cfg, mock_eab):
        """allowd_domain_list"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {
            "foo": "bar",
            "eab_handler": "handler",
            "eab_profiling": "eab",
        }
        mock_load_cfg.return_value = parser
        mock_eab.return_value = ["eab", "handler"]
        self.cahandler._config_load()
        self.assertFalse(self.cahandler.host)
        self.assertFalse(self.cahandler.user)
        self.assertFalse(self.cahandler.password)
        self.assertEqual("basic", self.cahandler.auth_method)
        self.assertFalse(self.cahandler.ca_bundle)
        self.assertFalse(self.cahandler.template)
        self.assertFalse(self.cahandler.krb5_config)
        self.assertEqual("handler", self.cahandler.eab_handler)
        self.assertEqual("eab", self.cahandler.eab_profiling)

    def test_030_revoke(self):
        """test revocation"""
        self.assertEqual(
            (
                500,
                "urn:ietf:params:acme:error:serverInternal",
                "Revocation is not supported.",
            ),
            self.cahandler.revoke("cert", "rev_reason", "rev_date"),
        )

    def test_031_poll(self):
        """test polling"""
        self.assertEqual(
            ("Method not implemented.", None, None, "poll_identifier", False),
            self.cahandler.poll("cert_name", "poll_identifier", "csr"),
        )

    def test_032_trigger(self):
        """test polling"""
        self.assertEqual(
            ("Method not implemented.", None, None), self.cahandler.trigger("payload")
        )

    def test_033_check_credentials(self):
        """test polling"""
        ca_server = Mock()
        ca_server.check_credentials = Mock(return_value=True)
        self.assertTrue(self.cahandler._check_credentials(ca_server))

    def test_034_check_credentials(self):
        """test polling"""
        ca_server = Mock()
        ca_server.check_credentials = Mock(return_value=False)
        self.assertFalse(self.cahandler._check_credentials(ca_server))

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.CAhandler._config_load")
    def test_035__enter__(self, mock_cfg):
        """test enter  called"""
        mock_cfg.return_value = True
        self.cahandler.__enter__()
        self.assertTrue(mock_cfg.called)

    def test_036_enroll(self):
        """enroll without having self.host"""
        self.assertEqual(
            ("Config incomplete", None, None, None), self.cahandler.enroll("csr")
        )

    def test_037_enroll(self):
        """enroll without having self.user"""
        self.cahandler.host = "host"
        self.assertEqual(
            ("Config incomplete", None, None, None), self.cahandler.enroll("csr")
        )

    def test_038_enroll(self):
        """enroll without having self.password"""
        self.cahandler.host = "host"
        self.cahandler.user = "user"
        self.assertEqual(
            ("Config incomplete", None, None, None), self.cahandler.enroll("csr")
        )

    def test_039_enroll(self):
        """enroll without having self.template"""
        self.cahandler.host = "host"
        self.cahandler.user = "user"
        self.cahandler.password = "password"
        self.assertEqual(
            ("Config incomplete", None, None, None), self.cahandler.enroll("csr")
        )

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.CAhandler._check_credentials")
    @patch("certsrv.Certsrv")
    def test_040_enroll(self, mock_certserver, mock_credchk):
        """enroll credential check failed"""
        self.cahandler.host = "host"
        self.cahandler.user = "user"
        self.cahandler.password = "password"
        self.cahandler.template = "template"
        mock_certserver.return_value = "foo"
        mock_credchk.return_value = False
        self.assertEqual(
            ("Connection or Credentialcheck failed.", None, None, None),
            self.cahandler.enroll("csr"),
        )

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.CAhandler._template_name_get")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.CAhandler._pkcs7_to_pem")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.convert_byte_to_string")
    @patch("textwrap.fill")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.CAhandler._check_credentials")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.Certsrv")
    def test_041_enroll(
        self, mock_certserver, mock_credchk, mockwrap, mock_b2s, mock_p2p, mock_tmpl
    ):
        """enroll enroll successful"""
        self.cahandler.host = "host"
        self.cahandler.user = "user"
        self.cahandler.password = "password"
        self.cahandler.template = "template"
        mockresponse = MagicMock()
        mockresponse.get_chain.return_value = "get_chain"
        mockresponse.get_cert.return_value = "get_cert"
        mock_certserver = mockresponse
        mock_credchk.return_value = True
        mockwrap.return_value = "mockwrap"
        mock_b2s.side_effect = ["get_chain", "get_cert"]
        mock_p2p.return_value = "p2p"
        self.assertEqual(
            (None, "get_certp2p", "get_cert", None), self.cahandler.enroll("csr")
        )
        self.assertFalse(mock_tmpl.called)

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.CAhandler._template_name_get")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.CAhandler._pkcs7_to_pem")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.convert_byte_to_string")
    @patch("textwrap.fill")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.CAhandler._check_credentials")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.Certsrv")
    def test_042_enroll(
        self,
        mock_certserver,
        mock_credchk,
        mockwrap,
        mock_b2s,
        mock_p2p,
        mock_tmpl,
    ):
        """enroll enroll successful"""
        self.cahandler.host = "host"
        self.cahandler.user = "user"
        self.cahandler.password = "password"
        self.cahandler.template = "template"
        mockresponse = MagicMock()
        mockresponse.get_chain.return_value = "get_chain"
        mockresponse.get_cert.return_value = "get_cert"
        mock_certserver = mockresponse
        mock_credchk.return_value = True
        mockwrap.return_value = "mockwrap"
        mock_b2s.side_effect = ["get_chain", "get_cert"]
        mock_p2p.return_value = "p2p"
        self.assertEqual(
            (None, "get_certp2p", "get_cert", None), self.cahandler.enroll("csr")
        )
        self.assertFalse(mock_tmpl.called)

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.CAhandler._template_name_get")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.CAhandler._pkcs7_to_pem")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.convert_byte_to_string")
    @patch("textwrap.fill")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.CAhandler._check_credentials")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.Certsrv")
    def test_043_enroll(
        self,
        mock_certserver,
        mock_credchk,
        mockwrap,
        mock_b2s,
        mock_p2p,
        mock_tmpl,
    ):
        """enroll enroll successful"""
        self.cahandler.host = "host"
        self.cahandler.user = "user"
        self.cahandler.password = "password"
        self.cahandler.template = "template"
        mockresponse = MagicMock()
        mockresponse.get_chain.return_value = "get_chain"
        mockresponse.get_cert.return_value = "get_cert"
        mock_certserver = mockresponse
        mock_credchk.return_value = True
        mockwrap.return_value = "mockwrap"
        mock_b2s.side_effect = ["get_chain", "get_cert"]
        mock_p2p.return_value = "p2p"
        self.assertEqual(
            (None, "get_certp2p", "get_cert", None), self.cahandler.enroll("csr")
        )
        self.assertFalse(mock_tmpl.called)

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.eab_profile_header_info_check")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.CAhandler._pkcs7_to_pem")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.convert_byte_to_string")
    @patch("textwrap.fill")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.CAhandler._check_credentials")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.Certsrv")
    def test_044_enroll(
        self, mock_certserver, mock_credchk, mockwrap, mock_b2s, mock_p2p, mock_eab
    ):
        """enroll enroll successful"""
        self.cahandler.host = "host"
        self.cahandler.user = "user"
        self.cahandler.password = "password"
        self.cahandler.template = "template"
        self.cahandler.header_info_field = "header_info"
        mockresponse = MagicMock()
        mockresponse.get_chain.return_value = "get_chain"
        mockresponse.get_cert.return_value = "get_cert"
        mock_eab.return_value = False
        mock_certserver = mockresponse
        mock_credchk.return_value = True
        mockwrap.return_value = "mockwrap"
        mock_b2s.side_effect = ["get_chain", "get_cert"]
        mock_p2p.return_value = "p2p"
        self.assertEqual(
            (None, "get_certp2p", "get_cert", None), self.cahandler.enroll("csr")
        )
        self.assertTrue(mock_eab.called)
        self.assertEqual("template", self.cahandler.template)

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.eab_profile_header_info_check")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.CAhandler._pkcs7_to_pem")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.convert_byte_to_string")
    @patch("textwrap.fill")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.CAhandler._check_credentials")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.Certsrv")
    def test_045_enroll(
        self, mock_certserver, mock_credchk, mockwrap, mock_b2s, mock_p2p, mock_eab
    ):
        """enroll enroll successful"""
        self.cahandler.host = "host"
        self.cahandler.user = "user"
        self.cahandler.password = "password"
        self.cahandler.template = "template"
        self.cahandler.header_info_field = "header_info"
        self.cahandler.krb5_config = "krb5_config"
        mockresponse = MagicMock()
        mockresponse.get_chain.return_value = "get_chain"
        mockresponse.get_cert.return_value = "get_cert"
        mock_eab.return_value = "error"
        mock_certserver = mockresponse
        mock_credchk.return_value = True
        mockwrap.return_value = "mockwrap"
        mock_b2s.side_effect = ["get_chain", "get_cert"]
        mock_p2p.return_value = "p2p"
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual(("error", None, None, None), self.cahandler.enroll("csr"))
        self.assertIn(
            "INFO:test_a2c:Load krb5config from krb5_config",
            lcm.output,
        )
        self.assertTrue(mock_eab.called)
        self.assertEqual("template", self.cahandler.template)

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.subprocess.run")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.os.path.isfile")
    def test_061a_kerberos_acquire_with_kinit_custom_path(
        self,
        mock_isfile,
        mock_subprocess_run,
    ):
        """test kinit fallback uses custom kinit binary and krb5 config"""
        self.cahandler.krb5_keytab = "/tmp/svc.keytab"
        self.cahandler.krb5_principal = "svc-a2c-enroll@EXAMPLE.COM"
        self.cahandler.krb5_config = "/tmp/krb5.conf"
        self.cahandler.krb5_kinit_path = "/usr/local/bin/kinit"

        mock_isfile.return_value = True
        result = self.cahandler._kerberos_acquire_with_kinit("/tmp/krb5cc_svc")

        self.assertTrue(result)
        run_args, run_kwargs = mock_subprocess_run.call_args
        self.assertEqual("/usr/local/bin/kinit", run_args[0][0])
        self.assertEqual("/tmp/krb5cc_svc", run_kwargs["env"]["KRB5CCNAME"])
        self.assertEqual("/tmp/krb5.conf", run_kwargs["env"]["KRB5_CONFIG"])

    def test_061b_credentials_are_configured_gssapi_keytab(self):
        """test gssapi keytab mode does not require user/password"""
        self.cahandler.auth_method = "gssapi"
        self.cahandler.krb5_principal = "svc-a2c-enroll@EXAMPLE.COM"
        self.cahandler.krb5_keytab = "/tmp/svc.keytab"

        self.assertTrue(self.cahandler._credentials_are_configured())

    def test_061c_enroll_gssapi_keytab_missing_template(self):
        """test enroll in gssapi keytab mode still needs template"""
        self.cahandler.host = "host"
        self.cahandler.auth_method = "gssapi"
        self.cahandler.krb5_principal = "svc-a2c-enroll@EXAMPLE.COM"
        self.cahandler.krb5_keytab = "/tmp/svc.keytab"

        self.assertEqual(
            ("Config incomplete", None, None, None), self.cahandler.enroll("csr")
        )

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.eab_profile_header_info_check")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.CAhandler._pkcs7_to_pem")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.convert_byte_to_string")
    @patch("textwrap.fill")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.CAhandler._check_credentials")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.Certsrv")
    def test_046_enroll(
        self, mock_certserver, mock_credchk, mockwrap, mock_b2s, mock_p2p, mock_eab
    ):
        """enroll enroll successful"""
        self.cahandler.host = "host"
        self.cahandler.user = "user"
        self.cahandler.password = "password"
        self.cahandler.template = "template"
        self.cahandler.header_info_field = "header_info"
        mockresponse = MagicMock()
        mockresponse.get_chain.return_value = "get_chain"
        mockresponse.get_cert.return_value = "get_cert"
        mock_eab.return_value = None
        mock_certserver = mockresponse
        mock_credchk.return_value = True
        mockwrap.return_value = "mockwrap"
        mock_b2s.side_effect = ["get_chain", "get_cert"]
        mock_p2p.return_value = "p2p"
        self.assertEqual(
            (None, "get_certp2p", "get_cert", None), self.cahandler.enroll("csr")
        )
        self.assertTrue(mock_eab.called)
        self.assertEqual("template", self.cahandler.template)

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.CAhandler._pkcs7_to_pem")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.convert_byte_to_string")
    @patch("textwrap.fill")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.CAhandler._check_credentials")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.Certsrv")
    def test_047_enroll(
        self, mock_certserver, mock_credchk, mockwrap, mock_b2s, mock_p2p
    ):
        """enroll exceütption in get chain"""
        self.cahandler.host = "host"
        self.cahandler.user = "user"
        self.cahandler.password = "password"
        self.cahandler.template = "template"
        mockresponse = MagicMock()
        mockresponse.get_chain.return_value = "get_chain"
        mockresponse.get_cert.return_value = "get_cert"
        mock_certserver = mockresponse
        mock_credchk.return_value = True
        mockwrap.return_value = "mockwrap"
        mock_b2s.side_effect = [Exception("exc_get_chain"), "get_cert"]
        mock_p2p.return_value = "p2p"
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual(
                (
                    "Certificate bundling failed: missing CA certificate or issued certificate.",
                    None,
                    "get_cert",
                    None,
                ),
                self.cahandler.enroll("csr"),
            )
        self.assertIn(
            "ERROR:test_a2c:Failed to get CA certificate chain: exc_get_chain",
            lcm.output,
        )
        self.assertIn(
            "ERROR:test_a2c:Failed to bundle certificates: missing ca_pem or cert_raw.",
            lcm.output,
        )

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.CAhandler._pkcs7_to_pem")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.convert_byte_to_string")
    @patch("textwrap.fill")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.CAhandler._check_credentials")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.Certsrv")
    def test_048_enroll(
        self, mock_certserver, mock_credchk, mockwrap, mock_b2s, mock_p2p
    ):
        """enroll exceütption in get cert"""
        self.cahandler.host = "host"
        self.cahandler.user = "user"
        self.cahandler.password = "password"
        self.cahandler.template = "template"
        mockresponse = MagicMock()
        mockresponse.get_chain.return_value = "get_chain"
        mockresponse.get_cert.return_value = "get_cert"
        mock_certserver = mockresponse
        mock_credchk.return_value = True
        mockwrap.return_value = "mockwrap"
        mock_b2s.side_effect = ["get_chain", Exception("get_cert")]
        mock_p2p.return_value = "p2p"
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual(
                ("get_cert", None, None, None), self.cahandler.enroll("csr")
            )
        self.assertIn(
            "ERROR:test_a2c:Failed to enroll certificate from CA: get_cert",
            lcm.output,
        )

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.enrollment_config_log")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.CAhandler._check_credentials")
    @patch("certsrv.Certsrv")
    def test_049_enroll(self, mock_certserver, mock_credchk, mock_ecl):
        """enroll credential check failed"""
        self.cahandler.host = "host"
        self.cahandler.user = "user"
        self.cahandler.password = "password"
        self.cahandler.template = "template"
        mock_certserver.return_value = "foo"
        mock_credchk.return_value = False
        self.assertEqual(
            ("Connection or Credentialcheck failed.", None, None, None),
            self.cahandler.enroll("csr"),
        )
        self.assertFalse(mock_ecl.called)

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.enrollment_config_log")
    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.CAhandler._check_credentials")
    @patch("certsrv.Certsrv")
    def test_050_enroll(self, mock_certserver, mock_credchk, mock_ecl):
        """enroll credential check failed"""
        self.cahandler.host = "host"
        self.cahandler.user = "user"
        self.cahandler.password = "password"
        self.cahandler.template = "template"
        self.cahandler.enrollment_config_log = True
        mock_certserver.return_value = "foo"
        mock_credchk.return_value = False
        self.assertEqual(
            ("Connection or Credentialcheck failed.", None, None, None),
            self.cahandler.enroll("csr"),
        )
        self.assertTrue(mock_ecl.called)

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.header_info_get")
    def test_051_template_name_get(self, mock_header):
        """test _template_name_get()"""
        mock_header.return_value = [
            {
                "header_info": '{"header_field": "template=foo lego-cli/4.14.2 xenolf-acme/4.14.2 (release; linux; amd64)"}'
            }
        ]
        self.cahandler.header_info_field = "header_field"
        self.assertEqual("foo", self.cahandler._template_name_get("csr"))

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.header_info_get")
    def test_052_template_name_get(self, mock_header):
        """test _template_name_get()"""
        mock_header.return_value = [
            {
                "header_info": '{"header_field": "Template=foo lego-cli/4.14.2 xenolf-acme/4.14.2 (release; linux; amd64)"}'
            }
        ]
        self.cahandler.header_info_field = "header_field"
        self.assertEqual("foo", self.cahandler._template_name_get("csr"))

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.header_info_get")
    def test_053_template_name_get(self, mock_header):
        """test _template_name_get()"""
        mock_header.return_value = [{"header_info": "header_info"}]
        self.cahandler.header_info_field = "header_field"
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertFalse(self.cahandler._template_name_get("csr"))
        self.assertIn(
            "ERROR:test_a2c:Failed to parse template from header_info: Expecting value: line 1 column 1 (char 0)",
            lcm.output,
        )

    def test_054_config_headerinfo_load(self):
        """test config_headerinfo_load()"""
        config_dic = {"Order": {"header_info_list": '["foo", "bar", "foobar"]'}}
        self.cahandler._config_headerinfo_load(config_dic)
        self.assertEqual("foo", self.cahandler.header_info_field)

    def test_055_config_headerinfo_load(self):
        """test config_headerinfo_load()"""
        config_dic = {"Order": {"header_info_list": '["foo"]'}}
        self.cahandler._config_headerinfo_load(config_dic)
        self.assertEqual("foo", self.cahandler.header_info_field)

    def test_056_config_headerinfo_load(self):
        """test config_headerinfo_load()"""
        config_dic = {"Order": {"header_info_list": "foo"}}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_headerinfo_load(config_dic)
        self.assertFalse(self.cahandler.header_info_field)
        self.assertIn(
            "WARNING:test_a2c:Failed to parse header_info_list from configuration: Expecting value: line 1 column 1 (char 0)",
            lcm.output,
        )

    def test_057__config_url_load(self):
        """test _config_url_load()"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"url": "foo"}
        self.cahandler._config_url_load(parser)
        self.assertEqual("foo", self.cahandler.url)

    @patch.dict("os.environ", {"url_variable": "foo1"})
    def test_058__config_url_load(self):
        """test _config_url_load()"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"url_variable": "url_variable"}
        self.cahandler._config_url_load(parser)
        self.assertEqual("foo1", self.cahandler.url)

    @patch.dict("os.environ", {"url_variable": "foo1"})
    def test_059__config_url_load(self):
        """test _config_url_load()"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"url_variable": "url_variable", "url": "foo"}
        self.cahandler._config_url_load(parser)
        self.assertEqual("foo", self.cahandler.url)

    @patch.dict("os.environ", {"url_variable": "foo1"})
    def test_060__config_url_load(self):
        """test _config_url_load()"""
        parser = configparser.ConfigParser()
        parser["CAhandler"] = {"url_variable": "doesnotexist"}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_url_load(parser)
        self.assertFalse(self.cahandler.url)
        self.assertIn(
            "ERROR:test_a2c:Could not load url_variable from environment: 'doesnotexist'",
            lcm.output,
        )

    @patch("acme2certifier.cahandlers.mscertsrv_ca_handler.handler_config_check")
    def test_061_handler_check(self, mock_handler_check):
        """test handler_check"""
        mock_handler_check.return_value = "mock_handler_check"
        self.assertEqual("mock_handler_check", self.cahandler.handler_check())


if __name__ == "__main__":

    unittest.main()



# -*- coding: utf-8 -*-
"""handler for an openssl ca"""

from __future__ import print_function
import os
import datetime
import json
from typing import List, Tuple, Dict
import base64
import uuid
import re
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.x509 import (
    BasicConstraints,
    ExtendedKeyUsage,
    SubjectKeyIdentifier,
    AuthorityKeyIdentifier,
    KeyUsage,
    SubjectAlternativeName,
)
from cryptography.x509.oid import ExtendedKeyUsageOID, NameOID

# pylint: disable=e0401
from acme2certifier.acme_srv.helper import (
    load_config,
    build_pem_file,
    uts_now,
    uts_to_date_utc,
    b64_url_recode,
    cert_serial_get,
    convert_string_to_byte,
    convert_byte_to_string,
    csr_cn_get,
    csr_san_get,
)

BLOCK_ALL_DOMAIN = "block.all"


class CAhandler(object):
    """CA  handler"""

    def __init__(self, debug: bool = False, logger: object = None):
        self.debug = debug
        self.logger = logger
        self.issuer_dict = {
            "issuing_ca_key": None,
            "issuing_ca_cert": None,
            "issuing_ca_crl": None,
        }
        self.ca_cert_chain_list = []
        self.cert_validity_days = 365
        self.cert_validity_adjust = False
        self.openssl_conf = None
        self.cert_save_path = None
        self.save_cert_as_hex = False
        self.allowed_domainlist = []
        self.blocked_domainlist = []
        self.cn_enforce = False

    def __enter__(self):
        """Makes ACMEHandler a Context Manager"""
        if not self.issuer_dict["issuing_ca_key"]:
            self._config_load()
        return self

    def __exit__(self, *args):
        """cose the connection at the end of the context"""

    def _ca_load(self) -> Tuple[object, object]:
        """load ca key and cert"""
        self.logger.debug("CAhandler._ca_load()")
        ca_key = None
        ca_cert = None
        # open key and cert
        if "issuing_ca_key" in self.issuer_dict:
            if os.path.exists(self.issuer_dict["issuing_ca_key"]):
                with open(self.issuer_dict["issuing_ca_key"], "rb") as fso:
                    ca_key = serialization.load_pem_private_key(
                        fso.read(),
                        password=self.issuer_dict.get("passphrase", None),
                        backend=default_backend(),
                    )
        if "issuing_ca_cert" in self.issuer_dict:
            if os.path.exists(self.issuer_dict["issuing_ca_cert"]):
                with open(self.issuer_dict["issuing_ca_cert"], "rb") as fso:
                    ca_cert = x509.load_pem_x509_certificate(
                        fso.read(), backend=default_backend()
                    )
        self.logger.debug("CAhandler._ca_load() ended")
        return (ca_key, ca_cert)

    def _cert_extension_ku_parse(self, ext: str) -> Dict[str, str]:
        self.logger.debug("CAhandler._cert_extension_ku_parse()")

        template_dic = {
            "digital_signature": False,
            "content_commitment": False,
            "key_encipherment": False,
            "data_encipherment": False,
            "key_agreement": False,
            "key_cert_sign": False,
            "crl_sign": False,
            "encipher_only": False,
            "decipher_only": False,
        }
        ku_mapping_dic = {
            "digitalsignature": "digital_signature",
            "nonrepudiation": "content_commitment",
            "keyencipherment": "key_encipherment",
            "dataencipherment": "data_encipherment",
            "keyagreement": "key_agreement",
            "keycertsign": "key_cert_sign",
            "crlsign": "crl_sign",
            "encipheronly": "encipher_only",
            "decipheronly": "decipher_only",
        }
        for attribute in ext.split(","):
            if attribute.strip().lower() in ku_mapping_dic:
                self.logger.debug(
                    "CAhandler._cert_extension_ku_parse(): found %s", attribute
                )
                template_dic[ku_mapping_dic[attribute.strip().lower()]] = True

        self.logger.debug("CAhandler._cert_extension_ku_parse() ended")
        return template_dic

    def _cert_extension_eku_parse(self, ext: str) -> List[str]:
        self.logger.debug("CAhandler._cert_extension_eku_parse()")

        # eku included in tempalate
        eku_mapping_dic = {
            "clientauth": ExtendedKeyUsageOID.CLIENT_AUTH,
            "serverauth": ExtendedKeyUsageOID.SERVER_AUTH,
            "codesigning": ExtendedKeyUsageOID.CODE_SIGNING,
            "emailprotection": ExtendedKeyUsageOID.EMAIL_PROTECTION,
            "timestamping": ExtendedKeyUsageOID.TIME_STAMPING,
            "ocspsigning": ExtendedKeyUsageOID.OCSP_SIGNING,
            "ekeyuse": "eKeyUse",  # this is just for testing
        }

        # backwards compatibility with cryptography module coming Ubuntu 22.04
        if hasattr(ExtendedKeyUsageOID, "KERBEROS_PKINIT_KDC"):
            eku_mapping_dic["pkInitKDC"] = ExtendedKeyUsageOID.KERBEROS_PKINIT_KDC

        eku_list = []
        for attribute in ext.split(","):
            if attribute.strip().lower() in eku_mapping_dic:
                self.logger.debug(
                    "CAhandler._cert_extension_eku_parse(): found %s", attribute
                )
                eku_list.append(eku_mapping_dic[attribute.strip().lower()])

        self.logger.debug("CAhandler._cert_extension_eku_parse() ended")
        return eku_list

    def _cert_extension_dic_parse(
        self, cert_extension_dic: Dict[str, str], cert: str, ca_cert: str
    ) -> List[object]:
        """parse certificate exteions loaded from config file"""
        self.logger.debug("CAhandler.cert_extesion_dic_parse()")

        extension_list = []
        for ext_name, ext in cert_extension_dic.items():
            _tmp_dic = {"critical": ext["critical"]}

            if ext_name.lower() == "basicconstraints":
                self.logger.debug(
                    "CAhandler.cert_extesion_dic_parse(): basicConstraints"
                )
                _tmp_dic["name"] = BasicConstraints(ca=False, path_length=None)
            elif ext_name.lower() == "subjectkeyidentifier":
                self.logger.debug(
                    "CAhandler.cert_extesion_dic_parse(): subjectKeyIdentifier"
                )
                _tmp_dic["name"] = SubjectKeyIdentifier.from_public_key(
                    cert.public_key()
                )
                _tmp_dic["critical"] = False
            elif ext_name.lower() == "authoritykeyidentifier":
                self.logger.debug(
                    "CAhandler.cert_extesion_dic_parse(): authorityKeyIdentifier"
                )
                _tmp_dic["name"] = AuthorityKeyIdentifier.from_issuer_public_key(
                    ca_cert.public_key()
                )
            elif ext_name.lower() == "keyusage":
                self.logger.debug("CAhandler.cert_extesion_dic_parse(): keyUsage")
                _tmp_dic["name"] = KeyUsage(
                    **self._cert_extension_ku_parse(ext["value"])
                )
            elif ext_name.lower() == "extendedkeyusage":
                self.logger.debug(
                    "CAhandler.cert_extesion_dic_parse(): extendedKeyUsage"
                )
                _tmp_dic["name"] = ExtendedKeyUsage(
                    self._cert_extension_eku_parse(ext["value"])
                )

            extension_list.append(_tmp_dic)

        self.logger.debug("CAhandler.cert_extesion_dic_parse() ended.")
        return extension_list

    def _certificate_extensions_load(self) -> Dict[str, str]:
        """verify certificate chain"""
        self.logger.debug("CAhandler._certificate_extensions_load()")

        file_dic = dict(load_config(self.logger, cfg_file=self.openssl_conf))

        cert_extention_dic = {}
        if "extensions" in file_dic:
            for extension in file_dic["extensions"]:

                cert_extention_dic[extension] = {}
                parameters = file_dic["extensions"][extension].split(",")

                # set crititcal task if applicable
                if parameters[0] == "critical":
                    cert_extention_dic[extension]["critical"] = bool(parameters.pop(0))
                else:
                    cert_extention_dic[extension]["critical"] = False

                # remove leading blank from first element
                parameters[0] = parameters[0].lstrip()

                # check if we have an issuer option (if so remove it and mark it as to be set)
                if "issuer:" in parameters[-1]:
                    cert_extention_dic[extension]["issuer"] = bool(parameters.pop(-1))

                # check if we have an issuer option (if so remove it and mark it as to be set)
                if "subject:" in parameters[-1]:
                    cert_extention_dic[extension]["subject"] = bool(parameters.pop(-1))

                # combine the remaining items and put them in as values
                cert_extention_dic[extension]["value"] = ",".join(parameters)

        self.logger.debug("CAhandler._certificate_extensions_load() ended")
        return cert_extention_dic

    def _certificate_store(self, cert: object):
        """store certificate on disk"""
        self.logger.debug("CAhandler._certificate_store()")
        serial = cert.serial_number

        # save cert if needed
        if self.cert_save_path and self.cert_save_path is not None:
            # create cert-store dir if not existing
            if not os.path.isdir(self.cert_save_path):
                self.logger.debug("create certsavedir %s", self.cert_save_path)
                os.mkdir(self.cert_save_path)

            # determine filename
            if self.save_cert_as_hex:
                self.logger.debug(
                    "Convert serial to hex: %s: %s", serial, f"{serial:X}"
                )
                cert_file = f"{serial:X}"
            else:
                cert_file = str(serial)
            with open(f"{self.cert_save_path}/{cert_file}.pem", "wb") as fso:
                fso.write(cert.public_bytes(serialization.Encoding.PEM))
        else:
            self.logger.error(
                "Certificate storage failed: cert_save_path is missing in the handler configuration."
            )

        self.logger.debug("CAhandler._certificate_store() ended")

    def _config_check_issuer(self) -> str:
        """check issuing CA configuration"""
        self.logger.debug("CAhandler._config_check_issuer()")

        error = None
        if "issuing_ca_key" in self.issuer_dict and self.issuer_dict["issuing_ca_key"]:
            if not os.path.exists(self.issuer_dict["issuing_ca_key"]):
                error = f"issuing_ca_key {self.issuer_dict['issuing_ca_key']} does not exist"
        else:
            error = "issuing_ca_key not specfied in config_file"

        if not error:
            if (
                "issuing_ca_cert" in self.issuer_dict
                and self.issuer_dict["issuing_ca_cert"]
            ):
                if not os.path.exists(self.issuer_dict["issuing_ca_cert"]):
                    error = f"issuing_ca_cert {self.issuer_dict['issuing_ca_cert']} does not exist"
            else:
                error = "issuing_ca_cert must be specified in config file"

        self.logger.debug("CAhandler._config_check_issuer() ended with: %s", error)
        return error

    def _config_check_crl(self, error: str = None) -> str:
        """check crl config"""
        self.logger.debug("CAhandler._config_check_crl()")

        if not error:
            if (
                "issuing_ca_crl" in self.issuer_dict
                and self.issuer_dict["issuing_ca_crl"]
            ):
                if not os.path.exists(self.issuer_dict["issuing_ca_crl"]):
                    self.logger.info(
                        "Issuing_ca_crl %s does not exist.",
                        self.issuer_dict["issuing_ca_crl"],
                    )
            else:
                error = "issuing_ca_crl must be specified in config file"

        self.logger.debug("CAhandler._config_check_crl() ended with: %s", error)
        return error

    def _config_parameters_check(self, error: str = None) -> str:
        """check remaining configuration"""
        self.logger.debug("CAhandler._config_parameters_check()")

        if not error:
            if self.cert_save_path:
                if not os.path.exists(self.cert_save_path):
                    error = f"cert_save_path {self.cert_save_path} does not exist"
            else:
                error = "cert_save_path must be specified in config file"

        if not error and self.openssl_conf and not os.path.exists(self.openssl_conf):
            error = f"openssl_conf {self.openssl_conf} does not exist"

        self.logger.debug("CAhandler._config_parameters_check() ended with: %s", error)
        return error

    def _config_check(self) -> str:
        """check config for consitency"""
        self.logger.debug("CAhandler._config_check()")

        # run checks
        error = self._config_check_issuer()
        error = self._config_check_crl(error)
        error = self._config_parameters_check(error)

        if error:
            self.logger.error("Configuration error: %s", error)

        self.logger.debug("CAhandler._config_check() ended")
        return error

    def _config_domainlists_load(self, config_dic: Dict[str, str]):
        """ " load config from file"""
        self.logger.debug("CAhandler._config_domainlists_load()")

        self.openssl_conf = config_dic.get(
            "CAhandler", "openssl_conf", fallback=self.openssl_conf
        )
        if "allowed_domainlist" in config_dic["CAhandler"]:
            try:
                self.allowed_domainlist = json.loads(
                    config_dic.get("CAhandler", "allowed_domainlist")
                )
            except Exception as err:
                self.logger.error(
                    "Unable to load allowed_domainlist parameter. Block all domains: %s",
                    err,
                )
                self.allowed_domainlist = [BLOCK_ALL_DOMAIN]

        if "blocked_domainlist" in config_dic["CAhandler"]:
            try:
                self.blocked_domainlist = json.loads(
                    config_dic.get("CAhandler", "blocked_domainlist")
                )
            except Exception as err:
                self.logger.error(
                    "Unable to load blocked_domainlist parameter. Block all domains: %s",
                    err,
                )
                self.allowed_domainlist = [BLOCK_ALL_DOMAIN]

        if "whitelist" in config_dic["CAhandler"]:
            self.logger.error(
                'Deprecated config: found "whitelist". Please rename to "allowed_domainlist".'
            )
            try:
                self.allowed_domainlist = json.loads(
                    config_dic.get("CAhandler", "whitelist")
                )
            except Exception as err:
                self.logger.error(
                    "Unable to load whitelist parameter. Block all domains: %s", err
                )
                self.allowed_domainlist = [BLOCK_ALL_DOMAIN]

        if "blacklist" in config_dic["CAhandler"]:
            self.logger.error(
                'Deprecated config: found "blacklist". Please rename to "blocked_domainlist".'
            )
            try:
                self.blocked_domainlist = json.loads(
                    config_dic.get("CAhandler", "blacklist")
                )

            except Exception as err:
                self.logger.error(
                    "Unable to load blacklist parameter. Block all domains: %s", err
                )
                self.allowed_domainlist = [BLOCK_ALL_DOMAIN]

        self.logger.debug("CAhandler._config_domainlists_load() ended")

    def _config_credentials_load(self, config_dic: Dict[str, str]):
        """load credential config"""
        self.logger.debug("CAhandler._config_credentials_load()")

        self.issuer_dict["issuing_ca_key"] = config_dic.get(
            "CAhandler", "issuing_ca_key", fallback=None
        )
        self.issuer_dict["issuing_ca_cert"] = config_dic.get(
            "CAhandler", "issuing_ca_cert", fallback=None
        )

        if "issuing_ca_key_passphrase_variable" in config_dic["CAhandler"]:
            try:
                self.issuer_dict["passphrase"] = os.environ[
                    config_dic.get("CAhandler", "issuing_ca_key_passphrase_variable")
                ]
            except Exception as err:
                self.logger.error(
                    "Unable to load issuing_ca_key_passphrase_variable from environment: %s",
                    err,
                )
        if "issuing_ca_key_passphrase" in config_dic["CAhandler"]:
            if "passphrase" in self.issuer_dict and self.issuer_dict["passphrase"]:
                self.logger.info("Overwrite issuing_ca_key_passphrase_variable")
            self.issuer_dict["passphrase"] = config_dic.get(
                "CAhandler", "issuing_ca_key_passphrase"
            )

        # convert passphrase
        if "passphrase" in self.issuer_dict:
            self.issuer_dict["passphrase"] = self.issuer_dict["passphrase"].encode(
                "ascii"
            )

        self.logger.debug("CAhandler._config_credentials_load() ended")

    def _config_policy_load(self, config_dic: Dict[str, str]):
        """load certificate policy"""
        self.logger.debug("CAhandler._config_policy_load()")

        self.cert_save_path = config_dic.get(
            "CAhandler", "cert_save_path", fallback=self.cert_save_path
        )
        self.issuer_dict["issuing_ca_crl"] = config_dic.get(
            "CAhandler", "issuing_ca_crl", fallback=None
        )

        if "ca_cert_chain_list" in config_dic["CAhandler"]:
            try:
                self.ca_cert_chain_list = json.loads(
                    config_dic["CAhandler"]["ca_cert_chain_list"]
                )
            except Exception as err:
                self.logger.error(
                    "Could not load ca_cert_chain_list parameter from config file: %s",
                    err,
                )
                self.ca_cert_chain = []

        if "cert_validity_days" in config_dic["CAhandler"]:
            self.cert_validity_days = int(config_dic["CAhandler"]["cert_validity_days"])
        try:
            self.cn_enforce = config_dic.getboolean(
                "CAhandler", "cn_enforce", fallback=False
            )
        except Exception:
            self.logger.error("Could not parse cn_enforce from config file.")
        try:
            self.cert_validity_adjust = config_dic.getboolean(
                "CAhandler", "cert_validity_adjust", fallback=False
            )
        except Exception:
            self.logger.error("Could not parse cert_validity_adjust from config file.")

        self.logger.debug("CAhandler._config_policy_load() ended")

    def _config_load(self):
        """ " load config from file"""
        self.logger.debug("CAhandler._config_load()")
        config_dic = load_config(self.logger, "CAhandler")

        # load credentials
        self._config_credentials_load(config_dic)

        # load policy options
        self._config_policy_load(config_dic)

        # load allow/block lists
        self._config_domainlists_load(config_dic)

        self.save_cert_as_hex = config_dic.getboolean(
            "CAhandler", "save_cert_as_hex", fallback=False
        )
        self.logger.debug("CAhandler._config_load() ended")

    def _chk_san_lists_get(self, csr: str) -> Tuple[List[str], List[bool]]:
        """check lists"""
        self.logger.debug("CAhandler._chk_san_lists_get()")

        # get sans and build a list
        _san_list = csr_san_get(self.logger, csr)

        check_list = []
        san_list = []

        if _san_list:
            for san in _san_list:
                try:
                    # SAN list must be modified/filtered)
                    _san_type, san_value = san.lower().split(":")
                    san_list.append(san_value)
                except Exception:
                    # force check to fail as something went wrong during parsing
                    check_list.append(False)
                    self.logger.debug(
                        "CAhandler._csr_check(): san_list parsing failed at entry: %s",
                        san,
                    )

        self.logger.debug("CAhandler._chk_san_lists_get() ended")
        return (san_list, check_list)

    def _cn_add(self, csr: str, san_list: List[str]) -> Tuple[List[str], str]:
        """add CN if required"""
        self.logger.debug("CAhandler._cn_add()")

        # get common name and attach it to san_list
        cn_ = csr_cn_get(self.logger, csr)

        if not cn_ and san_list:
            enforced_cn = san_list[0]
            self.logger.info("Enforce CN to %s", enforced_cn)
        else:
            enforced_cn = None

        if cn_:
            cn_ = cn_.lower()
            if cn_ not in san_list:
                # append cn to san_list
                self.logger.debug("Ahandler._csr_check(): append cn to san_list")
                san_list.append(cn_)

        self.logger.debug("CAhandler._cn_add() ended with: %s", enforced_cn)
        return (san_list, enforced_cn)

    def _csr_check(self, csr: str) -> Tuple[bool, str]:
        """check CSR against definied allowed_domainlists"""
        self.logger.debug("CAhandler._csr_check()")

        san_list, check_list = self._chk_san_lists_get(csr)
        san_list, enforced_cn = self._cn_add(csr, san_list)

        if self.allowed_domainlist or self.blocked_domainlist:
            result = False

            # go over the san list and check each entry
            for san in san_list:
                check_list.append(
                    self._string_wlbl_check(
                        san, self.allowed_domainlist, self.blocked_domainlist
                    )
                )

            if check_list:
                # cover a cornercase with empty checklist (no san, no cn)
                if False in check_list:
                    result = False
                else:
                    result = True
        else:
            result = True

        self.logger.debug(
            "CAhandler._csr_check() ended with: %s enforce_cn: %s", result, enforced_cn
        )
        return (result, enforced_cn)

    def _list_regex_check(self, entry: str, list_: List[str]) -> bool:
        """check entry against regex"""
        self.logger.debug("CAhandler._list_regex_check()")

        check_result = False
        for regex in list_:
            if regex.startswith("*."):
                regex = regex.replace("*.", ".")
            regex_compiled = re.compile(regex)
            if bool(regex_compiled.search(entry)):
                # parameter is in set flag accordingly and stop loop
                check_result = True

        self.logger.debug("CAhandler._list_regex_check() ended with: %s", check_result)
        return check_result

    def _list_check(self, entry: str, list_: List[str], toggle: bool = False) -> bool:
        """check string against list"""
        self.logger.debug("CAhandler._list_check(%s:%s)", entry, toggle)
        self.logger.debug("check against list: %s", str(list_))

        # default setting
        check_result = False

        if entry:
            if list_:
                check_result = self._list_regex_check(entry, list_)
            else:
                # empty list, flip parameter to make the check successful
                check_result = True

        if toggle:
            # toggle result if this is a blocked_domainlist
            check_result = not check_result

        self.logger.debug("CAhandler._list_check() ended with: %s", check_result)
        return check_result

    def _pemcertchain_generate(self, ee_cert: str, issuer_cert: str) -> str:
        """build pem chain"""
        self.logger.debug("CAhandler._pemcertchain_generate()")

        if issuer_cert:
            pem_chain = f"{ee_cert}{issuer_cert}"
        else:
            pem_chain = ee_cert

        for cert in self.ca_cert_chain_list:
            if os.path.exists(cert):
                with open(cert, "r", encoding="utf8") as fso:
                    cert_pem = fso.read()
                pem_chain = f"{pem_chain}{cert_pem}"

        self.logger.debug("CAhandler._pemcertchain_generate() ended")
        return pem_chain

    def _string_wlbl_check(
        self, entry: str, white_list: List[str], black_list: List[str]
    ) -> bool:
        """check single against allowed_domainlist and blocked_domainlist"""
        self.logger.debug("CAhandler._string_wlbl_check(%s)", entry)

        # default setting
        chk_result = False

        # check if entry is in white_list
        wl_check = self._list_check(entry, white_list)
        if wl_check:
            self.logger.debug("%s in white_list", entry)
            if black_list:
                # we need to check blocked_domainlist if there is a blocked_domainlist and wl check passed
                if self._list_check(entry, black_list):
                    self.logger.debug("%s in black_list", entry)
                else:
                    self.logger.debug("%s not in black_list", entry)
                    chk_result = True
            else:
                chk_result = wl_check
        else:
            self.logger.debug("%s not in white_list", entry)

        self.logger.debug(
            "CAhandler._string_wlbl_check(%s) ended with: %s", entry, chk_result
        )
        return chk_result

    def _cert_expiry_get(self, cert):
        """get expiry date of certificate"""
        self.logger.debug("CAhandler._cert_expiry_get()")

        expiry_date = cert.not_valid_after

        self.logger.debug("CAhandler._cert_expiry_get() ended")
        return expiry_date

    def _cacert_expiry_get(self):
        """get closesd expiry date of issuing CA"""
        self.logger.debug("CAhandler._cacert_expiry_get()")

        ca_list = self.ca_cert_chain_list
        if (
            self.issuer_dict["issuing_ca_cert"]
            and self.issuer_dict["issuing_ca_cert"] not in ca_list
        ):
            ca_list.append(self.issuer_dict["issuing_ca_cert"])

        expiry_days = 0
        cert = None

        for ca_cert in ca_list:
            if ca_cert:
                if os.path.exists(ca_cert):
                    with open(ca_cert, "rb") as fso:
                        ca_cert = x509.load_pem_x509_certificate(
                            fso.read(), backend=default_backend()
                        )
                        _tmp_expiry_days = (
                            self._cert_expiry_get(ca_cert) - datetime.datetime.now()
                        ).days
                        if not expiry_days or _tmp_expiry_days < expiry_days:
                            self.logger.debug(
                                "CAhandler._cacert_expiry_get(): set expiry_days to %s",
                                _tmp_expiry_days,
                            )
                            expiry_days = _tmp_expiry_days
                            cert = ca_cert
                else:
                    self.logger.error(
                        "CA file %s does not exist",
                        ca_cert,
                    )

        self.logger.debug("CAhandler._cacert_expiry_get() ended")
        return expiry_days, cert

    def _certexpiry_date_default(self) -> datetime.datetime:
        """set certificate validity"""
        self.logger.debug("CAhandler._certexpiry_date_default()")

        # default cert validity is taken from config
        cert_validity = datetime.datetime.now(
            datetime.timezone.utc
        ) + datetime.timedelta(days=self.cert_validity_days)

        self.logger.debug("CAhandler._certexpiry_date_default() ended")
        return cert_validity

    def _certexpiry_date_set(self) -> datetime.datetime:
        """set certificate validity"""
        self.logger.debug("CAhandler._certexpiry_date_set()")

        # default cert validity is taken from config
        cert_validity = self._certexpiry_date_default()

        if self.cert_validity_adjust:
            # adjust validity to match the validity of the issuing CA

            ca_cert_validity, cert = self._cacert_expiry_get()
            if ca_cert_validity < self.cert_validity_days:
                self.logger.info(
                    "Adjust validity to %s days.",
                    ca_cert_validity,
                )
                cert_validity = cert.not_valid_after

        self.logger.debug("CAhandler._certexpiry_date_set() ended")
        return cert_validity

    def _cert_signing_prep(self, ca_cert: object, req: object, subject: str) -> object:
        """enroll certificate"""
        # pylint: disable=R0914, R0915
        self.logger.debug("CAhandler._cert_signing_prep()")

        cert_validity = self._certexpiry_date_set()

        # sign csr
        builder = x509.CertificateBuilder()
        builder = builder.not_valid_before(datetime.datetime.now(datetime.timezone.utc))
        builder = builder.not_valid_after(cert_validity)
        builder = builder.issuer_name(ca_cert.subject)
        builder = builder.subject_name(subject)
        builder = builder.serial_number(uuid.uuid4().int)
        builder = builder.public_key(req.public_key())

        self.logger.debug("CAhandler._cert_signing_prep() ended")
        return builder

    def _cert_extension_default(self, ca_cert: object, req: object) -> List[str]:
        """add default extensions"""
        self.logger.debug("CAhandler._cert_extension_default()")

        default_extension_list = [
            {"name": BasicConstraints(ca=False, path_length=None), "critical": True},
            {
                "name": ExtendedKeyUsage(
                    [ExtendedKeyUsageOID.SERVER_AUTH, ExtendedKeyUsageOID.CLIENT_AUTH]
                ),
                "critical": False,
            },
            {
                "name": KeyUsage(
                    digital_signature=True,
                    content_commitment=False,
                    key_encipherment=True,
                    data_encipherment=False,
                    key_agreement=False,
                    key_cert_sign=False,
                    crl_sign=False,
                    encipher_only=False,
                    decipher_only=False,
                ),
                "critical": True,
            },
        ]
        if req:
            default_extension_list.append(
                {
                    "name": SubjectKeyIdentifier.from_public_key(req.public_key()),
                    "critical": False,
                },
            )
        if ca_cert:
            default_extension_list.append(
                {
                    "name": AuthorityKeyIdentifier.from_issuer_public_key(
                        ca_cert.public_key()
                    ),
                    "critical": False,
                }
            )

        self.logger.debug("CAhandler._cert_extension_default() ended")
        return default_extension_list

    def _cert_extension_apply(
        self, builder: object, ca_cert: object, req: object
    ) -> object:
        """add cert extensions"""
        self.logger.debug("CAhandler._cert_extension_apply()")

        # load certificate_profile (if applicable)
        if self.openssl_conf:
            cert_extension_dic = self._certificate_extensions_load()
            extension_list = self._cert_extension_dic_parse(
                cert_extension_dic, req, ca_cert
            )
        else:
            extension_list = self._cert_extension_default(ca_cert, req)

        # add subject alternative names
        if req:
            for ext in req.extensions:
                if ext.oid._name == "subjectAltName":  # pylint: disable=W0212
                    extension_list.append(
                        {"name": SubjectAlternativeName(ext.value), "critical": False}
                    )

        for extension in extension_list:
            # add extensions to csr
            builder = builder.add_extension(
                extension["name"], critical=extension["critical"]
            )

        self.logger.debug("CAhandler._cert_extension_apply() ended")
        return builder

    def enroll(self, csr: str) -> Tuple[str, str, str, str]:
        """enroll certificate"""
        # pylint: disable=R0914, R0915
        self.logger.debug("CAhandler.enroll()")

        cert_bundle = None
        cert_raw = None

        error = self._config_check()

        if not error:
            try:
                # check CN and SAN against black/whitlist
                result, enforce_cn = self._csr_check(csr)

                if result:
                    # prepare the CSR
                    csr = build_pem_file(
                        self.logger, None, b64_url_recode(self.logger, csr), None, True
                    )

                    # load ca cert and key
                    ca_key, ca_cert = self._ca_load()

                    # creating a rest from CSR
                    req = x509.load_pem_x509_csr(
                        convert_string_to_byte(csr), default_backend()
                    )
                    subject = req.subject

                    if self.cn_enforce and enforce_cn:
                        self.logger.info("Overwrite CN with %s", enforce_cn)
                        subject = x509.Name(
                            [x509.NameAttribute(NameOID.COMMON_NAME, enforce_cn)]
                        )

                    builder = self._cert_signing_prep(ca_cert, req, subject)
                    builder = self._cert_extension_apply(builder, ca_cert, req)

                    # sign certificate
                    cert = builder.sign(
                        private_key=ca_key,
                        algorithm=hashes.SHA256(),
                        backend=default_backend(),
                    )

                    # store certifiate
                    self._certificate_store(cert)
                    # create bundle and raw cert
                    with open(
                        self.issuer_dict["issuing_ca_cert"], "r", encoding="utf8"
                    ) as ca_fso:
                        cert_bundle = self._pemcertchain_generate(
                            convert_byte_to_string(
                                cert.public_bytes(serialization.Encoding.PEM)
                            ),
                            ca_fso.read(),
                        )
                        cert_raw = convert_byte_to_string(
                            base64.b64encode(
                                cert.public_bytes(serialization.Encoding.DER)
                            )
                        )
                else:
                    error = "urn:ietf:params:acme:badCSR"

            except Exception as err:
                self.logger.error(
                    "Certificate enrollment failed due to exception: %s", err
                )
                error = "Unknown exception"

        self.logger.debug("CAhandler.enroll() ended")
        return (error, cert_bundle, cert_raw, None)

    def poll(
        self, _cert_name: str, poll_identifier: str, _csr: str
    ) -> Tuple[str, str, str, str, bool]:
        """poll status of pending CSR and download certificates"""
        self.logger.debug("CAhandler.poll()")

        error = "Method not implemented."
        cert_bundle = None
        cert_raw = None
        rejected = False

        self.logger.debug("CAhandler.poll() ended")
        return (error, cert_bundle, cert_raw, poll_identifier, rejected)

    def _crlobject_build(
        self, ca_cert: object, serial: int
    ) -> Tuple[x509.CertificateRevocationListBuilder, object]:
        self.logger.debug("CAhandler._crlobject_build()")

        if os.path.exists(self.issuer_dict["issuing_ca_crl"]):
            self.logger.info(
                "Load existing crl %s)",
                self.issuer_dict["issuing_ca_crl"],
            )
            # load  existing CRL
            with open(self.issuer_dict["issuing_ca_crl"], "rb") as fso:
                crl_data = fso.read()
                crl = x509.load_pem_x509_crl(crl_data, default_backend())
            builder = x509.CertificateRevocationListBuilder()
            builder = builder.issuer_name(crl.issuer)
            # add crl certificates from file to the new crl object
            for revserial in crl:
                builder = builder.add_revoked_certificate(revserial)  # pragma: no cover

            # see if the cert to be revokek already in the list
            ret = crl.get_revoked_certificate_by_serial_number(serial)

        else:
            self.logger.info(
                "Create new crl %s)",
                self.issuer_dict["issuing_ca_crl"],
            )
            builder = x509.CertificateRevocationListBuilder()
            builder = builder.issuer_name(ca_cert.issuer)
            ret = None

        self.logger.debug("CAhandler._crlobject_build() ended")
        return (builder, ret)

    def revoke(
        self, cert_pem: str, rev_reason: str = "unspecified", rev_date: str = None
    ) -> Tuple[int, str, str]:
        """revoke certificate"""
        self.logger.debug("CAhandler.revoke(%s: %s)", rev_reason, rev_date)
        code = None
        message = None
        detail = None

        rev_date_format = "%y%m%d%H%M%SZ"

        # overwrite revocation date - we ignore what has been submitted
        rev_date = uts_to_date_utc(uts_now(), rev_date_format)

        if "issuing_ca_crl" in self.issuer_dict and self.issuer_dict["issuing_ca_crl"]:
            # load ca cert and key
            ca_key, ca_cert = self._ca_load()

            # turn of chain_check due to issues in pyopenssl (check is not working if key-usage is set)
            # result = self._certificate_chain_verify(cert, ca_cert)

            # get serial number from certicate to be revoked
            serial = cert_serial_get(self.logger, cert_pem)

            if ca_key and ca_cert and serial:

                # build crl object
                builder, ret = self._crlobject_build(ca_cert, serial)

                if not isinstance(ret, x509.RevokedCertificate):
                    # this is the revocation operation
                    # Set up the revoked entry
                    revoked_entry = (
                        x509.RevokedCertificateBuilder()
                        .serial_number(serial)
                        .revocation_date(
                            datetime.datetime.strptime(rev_date, rev_date_format)
                        )
                        .build(default_backend())
                    )
                    builder = builder.add_revoked_certificate(revoked_entry)

                    # Sign the CRL
                    crl = (
                        builder.last_update(
                            datetime.datetime.strptime(rev_date, rev_date_format)
                        )
                        .next_update(
                            datetime.datetime.strptime(rev_date, rev_date_format)
                        )
                        .sign(ca_key, hashes.SHA256())
                    )

                    # Save CRL
                    with open(self.issuer_dict["issuing_ca_crl"], "wb") as fso:
                        fso.write(crl.public_bytes(serialization.Encoding.PEM))
                    code = 200
                else:
                    code = 400
                    message = "urn:ietf:params:acme:error:alreadyRevoked"
                    detail = "Certificate has already been revoked"
            else:
                code = 400
                message = "urn:ietf:params:acme:error:serverInternal"
                detail = "configuration error"
        else:
            code = 400
            message = "urn:ietf:params:acme:error:serverInternal"
            detail = "Unsupported operation"

        self.logger.debug("CAhandler.revoke() ended")
        return (code, message, detail)

    def trigger(self, _payload: str) -> Tuple[str, str, str]:
        """process trigger message and return certificate"""
        self.logger.debug("CAhandler.trigger()")

        error = "Method not implemented."
        cert_bundle = None
        cert_raw = None

        self.logger.debug("CAhandler.trigger() ended with error: %s", error)
        return (error, cert_bundle, cert_raw)



# -*- coding: utf-8 -*-
"""ca handler for generic EST server"""

from __future__ import print_function
import os
import textwrap
import json
from typing import List, Tuple, Dict
import requests
from requests.auth import HTTPBasicAuth
from requests_pkcs12 import Pkcs12Adapter
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.serialization.pkcs7 import (
    load_pem_pkcs7_certificates,
    load_der_pkcs7_certificates,
)

# pylint: disable=e0401
from acme2certifier.acme_srv.helper import (
    load_config,
    b64_decode,
    b64_url_recode,
    convert_byte_to_string,
    convert_string_to_byte,
    parse_url,
    proxy_check,
    handler_config_check,
    pkcs7_to_pem,
)


class CAhandler(object):
    """EST CA  handler"""

    def __init__(self, _debug: bool = False, logger: object = None):
        self.logger = logger
        self.est_host = None
        self.est_client_cert = False
        self.cert_passphrase = False
        self.est_user = None
        self.est_password = None
        self.ca_bundle = True
        self.proxy = None
        self.request_timeout = 20
        self.session = None

    def __enter__(self):
        """Makes CAhandler a Context Manager"""
        if not self.est_host:
            self._config_load()
        return self

    def __exit__(self, *args):
        """cose the connection at the end of the context"""

    def _cacerts_get(self) -> Tuple[str, str]:
        """get ca certs from cerver"""
        self.logger.debug("CAhandler._cacerts_get()")
        error = None
        if self.est_host:
            try:
                if self.est_client_cert:
                    self.logger.debug("CAhandler._cacerts_get() by using client-certs")
                    # client auth
                    response = self.session.get(
                        self.est_host + "/cacerts",
                        verify=self.ca_bundle,
                        proxies=self.proxy,
                        timeout=self.request_timeout,
                    )
                else:
                    self.logger.debug(
                        "CAhandler._cacerts_get() by using userid/password"
                    )
                    response = self.session.get(
                        self.est_host + "/cacerts",
                        auth=HTTPBasicAuth(self.est_user, self.est_password),
                        verify=self.ca_bundle,
                        proxies=self.proxy,
                        timeout=self.request_timeout,
                    )
                pem = self._pkcs7_to_pem(b64_decode(self.logger, response.text))
            except Exception as err_:
                self.logger.error("Error while getting the CA certificates: %s", err_)
                error = err_
                pem = None
        else:
            self.logger.error(
                'Configuration incomplete: "est_host" parameter is missing'
            )
            error = None
            pem = None

        self.logger.debug("CAhandler._cacerts_get() ended with err: %s", error)
        return (error, pem)

    def _cert_bundle_create(
        self, error: str, ca_pem: str, cert_raw: str
    ) -> Tuple[str, str, str]:
        """create cert bundle"""
        self.logger.debug("CAhandler._cert_bundle_create()")

        cert_bundle = None
        if not error:
            cert_bundle = cert_raw + ca_pem
            cert_raw = cert_raw.replace("-----BEGIN CERTIFICATE-----\n", "")
            cert_raw = cert_raw.replace("-----END CERTIFICATE-----\n", "")
            cert_raw = cert_raw.replace("\n", "")
        else:
            self.logger.error("Simpleenroll error: %s", error)

        self.logger.debug("CAhandler._cert_bundle_create()")
        return (error, cert_bundle, cert_raw)

    def _config_host_load(self, config_dic: Dict[str, str]):
        """load est server address"""
        self.logger.debug("CAhandler._config_host_load()")

        if "est_host_variable" in config_dic["CAhandler"]:
            try:
                self.est_host = (
                    os.environ[config_dic.get("CAhandler", "est_host_variable")]
                    + "/.well-known/est"
                )
            except Exception as err:
                self.logger.error("Could not load est_host_variable:%s", err)
        if "est_host" in config_dic["CAhandler"]:
            if self.est_host:
                self.logger.info("Overwrite est_host")
            self.est_host = config_dic.get("CAhandler", "est_host") + "/.well-known/est"
        if not self.est_host:
            self.logger.error('Missing "est_host" parameter')

        self.logger.debug("CAhandler._config_host_load() ended")

    def _cert_passphrase_load(self, config_dic: Dict[str, str]):
        """load cert passphrase"""
        self.logger.debug("CAhandler._cert_passphrase_load()")
        if "cert_passphrase_variable" in config_dic["CAhandler"]:
            try:
                self.cert_passphrase = os.environ[
                    config_dic.get("CAhandler", "cert_passphrase_variable")
                ]
            except Exception as err:
                self.logger.error(
                    "Could not load cert_passphrase_variable:%s",
                    err,
                )
        if "cert_passphrase" in config_dic["CAhandler"]:
            if self.cert_passphrase:
                self.logger.info("Overwrite cert_passphrase")
            self.cert_passphrase = config_dic.get("CAhandler", "cert_passphrase")
        self.logger.debug("CAhandler._cert_passphrase_load() ended")

    def _config_clientauth_load(self, config_dic: Dict[str, str]):
        """check if we need to use clientauth"""
        self.logger.debug("CAhandler._config_clientauth_load()")

        # client auth via pem files
        if "est_client_cert" in config_dic["CAhandler"]:
            if "est_client_key" in config_dic["CAhandler"]:
                self.logger.debug("CAhandler._config_clientauth_load(): load pem")
                self.est_client_cert = config_dic.get(
                    "CAhandler", "est_client_cert", fallback=self.est_client_cert
                )
                self.session.cert = (
                    config_dic.get("CAhandler", "est_client_cert"),
                    config_dic.get("CAhandler", "est_client_key"),
                )
            elif (
                "cert_passphrase" in config_dic["CAhandler"]
                or "cert_passphrase_variable" in config_dic["CAhandler"]
            ):
                self.logger.debug("CAhandler._config_clientauth_load(): load pkcs12")
                self.est_client_cert = config_dic.get("CAhandler", "est_client_cert")
                self._cert_passphrase_load(config_dic)
                self.session.mount(
                    self.est_host,
                    Pkcs12Adapter(
                        pkcs12_filename=config_dic.get("CAhandler", "est_client_cert"),
                        pkcs12_password=self.cert_passphrase,
                    ),
                )
            else:
                self.logger.error(
                    'Clientauth configuration incomplete: either "est_client_key or "cert_passphrase" parameter is missing in config file'
                )

        self.logger.debug("CAhandler._config_clientauth_load() ended")

    def _config_userauth_load(self, config_dic: Dict[str, str]):
        """check if we need to use user-auth"""
        self.logger.debug("CAhandler._config_userauth_load()")

        if "est_user_variable" in config_dic["CAhandler"]:
            try:
                self.est_user = os.environ[
                    config_dic.get("CAhandler", "est_user_variable")
                ]
            except Exception as err:
                self.logger.error("Could not load est_user_variable:%s", err)
        if "est_user" in config_dic["CAhandler"]:
            if self.est_user:
                self.logger.info("CAhandler._config_load() overwrite est_user")
            self.est_user = config_dic.get("CAhandler", "est_user")

        self.logger.debug("CAhandler._config_userauth_load() ended")

    def _config_password_load(self, config_dic: Dict[str, str]):
        """load password"""
        self.logger.debug("CAhandler._config_password_load()")

        if "est_password_variable" in config_dic["CAhandler"]:
            try:
                self.est_password = os.environ[
                    config_dic.get("CAhandler", "est_password_variable")
                ]
            except Exception as err:
                self.logger.error("Could not load est_password:%s", err)
        if "est_password" in config_dic["CAhandler"]:
            if self.est_password:
                self.logger.info("Overwrite est_password")
            self.est_password = config_dic.get("CAhandler", "est_password")

        if (self.est_user and not self.est_password) or (
            self.est_password and not self.est_user
        ):
            self.logger.error(
                'Configuration incomplete: either "est_user" or "est_password" parameter is missing in config file'
            )

        self.logger.debug("CAhandler._config_password_load() ended")

    def _config_parameters_load(self, config_dic: Dict[str, str]):
        """load config paramters"""
        self.logger.debug("CAhandler._config_load()")

        # check if we get a ca bundle for verification
        try:
            self.ca_bundle = config_dic.getboolean("CAhandler", "ca_bundle")
        except Exception:
            self.ca_bundle = config_dic.get(
                "CAhandler", "ca_bundle", fallback=self.ca_bundle
            )

        try:
            self.request_timeout = int(
                config_dic.get(
                    "CAhandler", "request_timeout", fallback=self.request_timeout
                )
            )
        except Exception:
            self.logger.error(
                "Could not load request_timeout:%s",
                config_dic.get("CAhandler", "request_timeout"),
            )

        self.logger.debug("CAhandler._config_load() ended")

    def _config_proxy_load(self, config_dic: Dict[str, str]):
        """load config paramters"""
        self.logger.debug("CAhandler._config_proxy_load()")

        if "DEFAULT" in config_dic and "proxy_server_list" in config_dic["DEFAULT"]:
            try:
                proxy_list = json.loads(config_dic.get("DEFAULT", "proxy_server_list"))
                url_dic = parse_url(self.logger, self.est_host)
                if "host" in url_dic:
                    fqdn, _port = url_dic["host"].split(":")
                    proxy_server = proxy_check(self.logger, fqdn, proxy_list)
                    self.proxy = {"http": proxy_server, "https": proxy_server}
            except Exception as err_:
                self.logger.warning(
                    "Failed to load proxy_server_list from configuration: %s",
                    err_,
                )

        self.logger.debug("CAhandler._config_proxy_load() ended")

    def _config_load(self):
        """ " load config from file"""
        # pylint: disable=R0912, R0915
        self.logger.debug("CAhandler._config_load()")
        config_dic = load_config(self.logger, "CAhandler")

        if "CAhandler" in config_dic:

            with requests.Session() as self.session:
                # load host information
                self._config_host_load(config_dic)
                # load clientauth
                self._config_clientauth_load(config_dic)
                # load user
                self._config_userauth_load(config_dic)
                # load password
                self._config_password_load(config_dic)
                # load paramters
                self._config_parameters_load(config_dic)
                # check if we have one authentication scheme
                if not self.est_client_cert and not self.est_user:
                    self.logger.error(
                        "Configuration incomplete: either user or client authentication must be configured."
                    )
                elif self.est_client_cert and self.est_user:
                    self.logger.error(
                        "Configuration error: user and client authentication cannot be configured together."
                    )
                if self.est_client_cert and not self.ca_bundle:
                    self.logger.error(
                        "Configuration error: client authentication requires a ca_bundle."
                    )

        # load proxy information
        self._config_proxy_load(config_dic)

        self.logger.debug("CAhandler._config_load() ended")

    def _pkcs7_to_pem(self, pkcs7_content: str, outform: str = "string") -> List[str]:
        """convert pkcs7 to pem"""
        self.logger.debug("CAhandler._pkcs7_to_pem()")

        result = pkcs7_to_pem(self.logger, pkcs7_content, outform)

        self.logger.debug("Certificate._pkcs7_to_pem() ended")
        return result

    def _simpleenroll(self, csr: str) -> Tuple[str, str]:
        """EST /simpleenroll request."""
        self.logger.debug("CAhandler._simpleenroll()")
        error = None
        try:
            headers = {"Content-Type": "application/pkcs10"}
            if self.est_client_cert:
                # client auth
                response = self.session.post(
                    self.est_host + "/simpleenroll",
                    data=csr,
                    headers=headers,
                    verify=self.ca_bundle,
                    proxies=self.proxy,
                    timeout=self.request_timeout,
                )
            else:
                response = self.session.post(
                    self.est_host + "/simpleenroll",
                    data=csr,
                    auth=HTTPBasicAuth(self.est_user, self.est_password),
                    headers=headers,
                    verify=self.ca_bundle,
                    proxies=self.proxy,
                    timeout=self.request_timeout,
                )
            # response.raise_for_status()
            pem = self._pkcs7_to_pem(b64_decode(self.logger, response.text))
        except Exception as err_:
            self.logger.error("Enrollment error: %s", err_)
            error = str(err_)
            pem = None

        self.logger.debug("CAhandler._simpleenroll() ended with err: %s", error)
        return (error, pem)

    def enroll(self, csr: str) -> Tuple[str, str, str, bool]:
        """enroll certificate from NCLM"""
        self.logger.debug("CAhandler.enroll()")
        cert_bundle = None
        error = None
        cert_raw = None

        # recode csr
        csr = textwrap.fill(b64_url_recode(self.logger, csr), 64) + "\n"

        if self.est_host:
            error, ca_pem = self._cacerts_get()

        if not error:
            if ca_pem:
                error, cert_raw = self._simpleenroll(csr)
                # build certificate bundle
                error, cert_bundle, cert_raw = self._cert_bundle_create(
                    error, ca_pem, cert_raw
                )
            else:
                error = "no CA certificates found"
                self.logger.error("No CA certificates found")
        else:
            self.logger.error("Error while fetching the CA certificates: %s", error)

        self.logger.debug("Certificate.enroll() ended")
        return (error, cert_bundle, cert_raw, None)

    def handler_check(self):
        """check if handler is ready"""
        self.logger.debug("CAhandler.check()")
        error = handler_config_check(self.logger, self, ["est_host"])
        self.logger.debug("CAhandler.check() ended with %s", error)
        return error

    def poll(
        self, _cert_name: str, poll_identifier: str, _csr: str
    ) -> Tuple[str, str, str, str, bool]:
        """poll status of pending CSR and download certificates"""
        self.logger.debug("CAhandler.poll()")

        error = "Method not implemented."
        cert_bundle = None
        cert_raw = None
        rejected = False

        self.logger.debug("CAhandler.poll() ended")
        return (error, cert_bundle, cert_raw, poll_identifier, rejected)

    def revoke(
        self, _cert: str, _rev_reason: str, _rev_date: str
    ) -> Tuple[int, str, str]:
        """revoke certificate"""
        self.logger.debug("CAhandler.tsg_id_lookup()")

        code = 500
        message = "urn:ietf:params:acme:error:serverInternal"
        detail = "Revocation is not supported."

        self.logger.debug("CAhandler.revoke() ended")
        return (code, message, detail)

    def trigger(self, _payload: str) -> Tuple[str, str, str]:
        """process trigger message and return certificate"""
        self.logger.debug("CAhandler.trigger()")

        error = "Method not implemented."
        cert_bundle = None
        cert_raw = None

        self.logger.debug("CAhandler.trigger() ended with error: %s", error)
        return (error, cert_bundle, cert_raw)



"""acme2certifier URL Configuration"""

from django.urls import include, re_path
from django.contrib import admin
from django.views.generic import RedirectView
from acme2certifier.django_app import views
from acme2certifier.acme_srv.helper import load_config

CONFIG = load_config()

if "Directory" in CONFIG and "url_prefix" in CONFIG["Directory"]:
    PREFIX = CONFIG["Directory"]["url_prefix"] + "/"
    if PREFIX.startswith("/"):
        PREFIX = PREFIX.lstrip("/")
else:
    PREFIX = ""

urlpatterns = [
    re_path(r"^admin/", admin.site.urls),
    re_path(r"^$", RedirectView.as_view(url="/directory")),
    re_path(r"^directory$", views.directory, name="directory"),
    re_path(rf"^{PREFIX}get_servername$", views.servername_get, name="servername_get"),
    re_path(rf"^{PREFIX}trigger$", views.trigger, name="trigger"),
    re_path(rf"^{PREFIX}housekeeping$", views.housekeeping, name="housekeeping"),
    re_path(rf"^{PREFIX}acme/", include("acme2certifier.django_app.urls")),
]

if "CAhandler" in CONFIG and "acme_url" in CONFIG["CAhandler"]:
    urlpatterns.append(
        re_path(
            rf"^{PREFIX}.well-known/acme-challenge/",
            views.acmechallenge_serve,
            name="acmechallenge_serve",
        )
    )
